/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                             reLOC 0.06-hanshin                             */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// cnf.h / 0.06-hanshin_029
/*----------------------------------------------------------------------------*/
//
// Dimacs CNF formula production tools.
//
/*----------------------------------------------------------------------------*/


#ifndef __CNF_H__
#define __CNF_H__

#include <vector>
#include <map>

#include "types.h"
#include "defs.h"
#include "result.h"


using namespace std;
using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/

    class sIndexScope;

    class sIndexableIdentifier;
    class sSpecifiedIdentifier;

    class sIndexableStateIdentifier;
    class sSpecifiedStateIdentifier;


/*----------------------------------------------------------------------------*/
// sIndex

    class sIndex
    {
    public:
	virtual sIndex* clone(void) const = 0;

	virtual void increment(void) = 0;
	virtual void decrement(void) = 0;

	virtual void increment(int distance) = 0;
	virtual void decrement(int distance) = 0;

	virtual bool operator==(const sIndex &index) const = 0;
	virtual bool operator!=(const sIndex &index) const = 0;

	virtual bool operator<=(const sIndex &index) const = 0;
	virtual bool operator<(const sIndex &index) const = 0;
	virtual int operator-(const sIndex &index) const = 0;

	virtual bool operator>=(const sIndex &index) const;
	virtual bool operator>(const sIndex &index) const;

	virtual sString to_String(void) const = 0;
	virtual void to_Screen(void) const = 0;
    };


/*----------------------------------------------------------------------------*/
// sIntegerIndex

    class sIntegerIndex
    : public sIndex
    {
    public:
	sIntegerIndex();
	sIntegerIndex(int value);
	sIntegerIndex(const sIntegerIndex &integer_index);
	const sIntegerIndex& operator=(const sIntegerIndex &integer_index);

	virtual sIndex* clone(void) const;

	int get_Value(void) const;
	void set_Value(int value);

	virtual void increment(void);
	virtual void decrement(void);

	virtual void increment(int distance);
	virtual void decrement(int distance);

	virtual bool operator==(const sIndex &index) const;
	virtual bool operator!=(const sIndex &index) const;

	virtual bool operator<=(const sIndex &index) const;
	virtual bool operator<(const sIndex &index) const;
	virtual int operator-(const sIndex &index) const;

	virtual sString to_String(void) const;
	virtual void to_Screen(void) const;

    private:
	int m_value;
    };


/*----------------------------------------------------------------------------*/
// sIndex_iterator

    class sIndex_iterator
    {
    public:
	sIndex_iterator(const sIndex &index);
	sIndex_iterator(const sIndex_iterator &index_iterator);
	const sIndex_iterator& operator=(const sIndex_iterator &index_iterator);
	~sIndex_iterator();

	sIndex_iterator& operator++();
	sIndex_iterator& operator--();

	sIndex_iterator& operator+=(int distance);
	sIndex_iterator& operator-=(int distance);

	bool operator==(const sIndex_iterator &iterator) const;
	bool operator!=(const sIndex_iterator &iterator) const;

	const sIndex& operator*() const;

    private:
	sIndex *m_index;
    };


/*----------------------------------------------------------------------------*/
// sIndexScope

    class sIndexScope
    {
    public:
	sIndexScope();
	sIndexScope(const sIndexScope &scope);
	const sIndexScope& operator=(const sIndexScope &scope);

	virtual sIndexScope* clone(void) const = 0;
	virtual int get_Size(void) const = 0;

	virtual sIndex_iterator get_Begin(void) const = 0;
	virtual sIndex_iterator get_End(void) const = 0;
    };


/*----------------------------------------------------------------------------*/
// sIntegerScope

    class sIntegerScope
      : public sIndexScope
    {
    public:
	sIntegerScope();
	sIntegerScope(int min, int max);
	sIntegerScope(const sIntegerScope &integer_scope);
	const sIntegerScope& operator=(const sIntegerScope &integer_scope);

	virtual sIndexScope* clone(void) const;
	virtual int get_Size(void) const;
 
	virtual sIndex_iterator get_Begin(void) const;
	virtual sIndex_iterator get_End(void) const;

	inline int get_Min(void) const;
	inline int get_Max(void) const;

	inline void set_Min(int min);
	inline void set_Max(int max);

    private:
	int m_min;
	int m_max;
    };


/*----------------------------------------------------------------------------*/

    inline int sIntegerScope::get_Min(void) const
    {
	return m_min;
    }

    
    inline int sIntegerScope::get_Max(void) const
    {
	return m_max;
    }


    inline void sIntegerScope::set_Min(int min)
    {
	m_min = min;
    }


    inline void sIntegerScope::set_Max(int max)
    {
	m_max = max;
    }


/*----------------------------------------------------------------------------*/
// sVariableStore_CNF

    class sVariableStore_CNF
    {
    public:
	sVariableStore_CNF();
	sVariableStore_CNF(const sVariableStore_CNF &variable_store);
	const sVariableStore_CNF& operator=(const sVariableStore_CNF &variable_store);

	int get_Last_CNFVariable(void) const;
	void alloc_CNFVariables(int N_Variables);

    private:
	int m_last_variable;
    };


/*----------------------------------------------------------------------------*/
// sSpecifiedIdentifier

    class sSpecifiedIdentifier
    {
    public:
	typedef std::vector<const sIndex*> Indexes_vector;

    public:
	sSpecifiedIdentifier();
	sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier);
	sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const sIndex &index);
	sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const sIndex &index_1, const sIndex &index_2);
	sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const sIndex &index_1, const sIndex &index_2, const sIndex &index_3);
	sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const Indexes_vector &scope_Indexes);

	sSpecifiedIdentifier(const sSpecifiedIdentifier &specified_identifier);
	const sSpecifiedIdentifier& operator=(const sSpecifiedIdentifier &specified_identifier);
	~sSpecifiedIdentifier();

	inline bool is_Null(void) const;
	inline const sIndexableIdentifier* get_IndexableIdentifier(void) const;
	inline Indexes_vector& get_ScopeIndexes(void);

	sString calc_String(void) const;
	int calc_CNF(void) const;

    private:
	const sIndexableIdentifier *m_indexable_identifier;
	Indexes_vector m_scope_Indexes;
    };


/*----------------------------------------------------------------------------*/

    inline bool sSpecifiedIdentifier::is_Null(void) const
    {
	return (m_indexable_identifier == NULL);
    }


    inline const sIndexableIdentifier* sSpecifiedIdentifier::get_IndexableIdentifier(void) const
    {
	return m_indexable_identifier;
    }


    inline sSpecifiedIdentifier::Indexes_vector& sSpecifiedIdentifier::get_ScopeIndexes(void)
    {
	return m_scope_Indexes;
    }


/*----------------------------------------------------------------------------*/
// sIndexableIdentifier

    class sIndexableIdentifier
    {
    public:
	typedef sSpecifiedIdentifier::Indexes_vector Indexes_vector;
	typedef std::vector<const sIndexScope*> IndexScopes_vector;

    public:
	sIndexableIdentifier();
	sIndexableIdentifier(sVariableStore_CNF *variable_store, const sString &base_name);
	sIndexableIdentifier(sVariableStore_CNF *variable_store, const sString &base_name, const sIndexScope &scope);

	sIndexableIdentifier(sVariableStore_CNF *variable_store,
			   const sString      &base_name,
			   const sIndexScope  &scope_1,
			   const sIndexScope  &scope_2);

	sIndexableIdentifier(sVariableStore_CNF *variable_store,
			   const sString      &base_name,
			   const sIndexScope  &scope_1,
			   const sIndexScope  &scope_2,
			   const sIndexScope  &scope_3);

	sIndexableIdentifier(sVariableStore_CNF       *variable_store,
			   const sString             &base_name,
			   const IndexScopes_vector &index_Scopes);

	sIndexableIdentifier(const sIndexableIdentifier &indexable_identifier);
	const sIndexableIdentifier& operator=(const sIndexableIdentifier &indexable_identifier);
	~sIndexableIdentifier();

	inline bool is_Anonymous(void) const;
	inline const sString& get_BaseName(void) const;

	inline int get_Arity(void) const;
	inline int get_First_CNFVariable(void) const;
	inline int get_CNFVariable_Count(void) const;
	
	inline const sIndexScope& get_IndexScope(int scope) const;
	inline const IndexScopes_vector& get_IndexScopes(void) const;

	sString calc_String(void) const;
	sString calc_String(const sIndex &index) const;
	sString calc_String(const sIndex &index_1, const sIndex &index_2) const;
	sString calc_String(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3) const;
	sString calc_String(const Indexes_vector &scope_Indexes) const;

	int calc_CNF(void) const;
	int calc_CNF(const sIndex &index) const;
	int calc_CNF(const sIndex &index_1, const sIndex &index_2) const;
	int calc_CNF(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3) const;
	int calc_CNF(const Indexes_vector &scope_Indexes) const;

	sSpecifiedIdentifier translate_CNFVariable(int cnf_variable) const;

	virtual void to_Screen(const sString &indent = "") const;

    private:
	virtual void to_Screen_identifiers(const Indexes_vector &scope_Indexes, const sString &indent = "") const;

    private:
	sVariableStore_CNF *m_variable_store;
	sString m_base_name;

	int m_first_cnf_variable;
	int m_N_cnf_Variables;

	IndexScopes_vector m_index_Scopes;
    };


/*----------------------------------------------------------------------------*/

    inline bool sIndexableIdentifier::is_Anonymous(void) const
    {
	return (m_variable_store == NULL);
    }

    inline const sString& sIndexableIdentifier::get_BaseName(void) const
    {
	return m_base_name;
    }


    inline int sIndexableIdentifier::get_Arity(void) const
    {
	return m_index_Scopes.size();
    }


    inline int sIndexableIdentifier::get_First_CNFVariable(void) const
    {
	return m_first_cnf_variable;
    }


    inline int sIndexableIdentifier::get_CNFVariable_Count(void) const
    {
	return m_N_cnf_Variables;
    }


    inline const sIndexScope& sIndexableIdentifier::get_IndexScope(int scope) const
    {
	sASSERT(get_Arity() > scope);
	return *m_index_Scopes[scope];
    }


    inline  const sIndexableIdentifier::IndexScopes_vector& sIndexableIdentifier::get_IndexScopes(void) const
    {
	return m_index_Scopes;
    }


/*----------------------------------------------------------------------------*/
// sSpecifiedStateIdentifier

    class sSpecifiedStateIdentifier
    {
    public:
	typedef sIndexableIdentifier::Indexes_vector Indexes_vector;

    public:
	sSpecifiedStateIdentifier();
	sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier);
	sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const sIndex &index);
	sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const sIndex &index_1, const sIndex &index_2);
	sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const sIndex &index_1, const sIndex &index_2, const sIndex &index_3);
	sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const Indexes_vector &scope_Indexes);

	sSpecifiedStateIdentifier(const sSpecifiedStateIdentifier &specified_identifier);
	const sSpecifiedStateIdentifier& operator=(const sSpecifiedStateIdentifier &specified_identifier);
	~sSpecifiedStateIdentifier();

	inline bool is_Null(void) const;
	inline const sIndexableStateIdentifier* get_StateIdentifier(void) const;

	sString calc_String(int bit) const;
	int calc_CNF(int bit) const;

    private:
	const sIndexableStateIdentifier *m_state_identifier;
	Indexes_vector m_scope_Indexes;
    };


/*----------------------------------------------------------------------------*/

    inline bool sSpecifiedStateIdentifier::is_Null(void) const
    {
	return (m_state_identifier == NULL);
    }


    inline const sIndexableStateIdentifier* sSpecifiedStateIdentifier::get_StateIdentifier(void) const
    {
	return m_state_identifier;
    }


/*----------------------------------------------------------------------------*/
// sIndexableStateIdentifier

    class sIndexableStateIdentifier
    {
	friend class sClauseGenerator;

    public:
	typedef sSpecifiedStateIdentifier::Indexes_vector Indexes_vector;
	typedef sIndexableIdentifier::IndexScopes_vector IndexScopes_vector;

    public:
	sIndexableStateIdentifier();
	sIndexableStateIdentifier(sVariableStore_CNF *variable_store, const sString &base_name, int N_States);
	sIndexableStateIdentifier(sVariableStore_CNF *variable_store, const sString &base_name, int N_States, const sIndexScope &scope);

	sIndexableStateIdentifier(sVariableStore_CNF *variable_store,
				const sString      &base_name,
				int                 N_States,
				const sIndexScope  &scope_1,
				const sIndexScope  &scope_2);

	sIndexableStateIdentifier(sVariableStore_CNF *variable_store,
				const sString      &base_name,
				int                 N_States,
				const sIndexScope  &scope_1,
				const sIndexScope  &scope_2,
				const sIndexScope  &scope_3);

	sIndexableStateIdentifier(sVariableStore_CNF       *variable_store,
				const sString            &base_name,
				int                       N_States,
				const IndexScopes_vector &index_Scopes);

	sIndexableStateIdentifier(const sIndexableStateIdentifier &state_identifier);
	const sIndexableStateIdentifier& operator=(const sIndexableStateIdentifier &state_identifier);
	~sIndexableStateIdentifier();

	inline int get_Arity(void) const;
	inline int get_First_CNFVariable(void) const;
	inline int get_CNFVariable_Count(void) const;

	inline int get_StateCount(void) const;
	inline int get_Log2_StateCount(void) const;

	inline const sIndexableIdentifier& get_IndexableIdentifier(void) const;

	sString calc_String(int bit) const;
	sString calc_String(const sIndex &index, int bit) const;
	sString calc_String(const sIndex &index_1, const sIndex &index_2, int bit) const;
	sString calc_String(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3, int bit) const;
	sString calc_String(const Indexes_vector &scope_Indexes, int bit) const;

	int calc_CNF(int bit) const;
	int calc_CNF(const sIndex &index, int bit) const;
	int calc_CNF(const sIndex &index_1, const sIndex &index_2, int bit) const;
	int calc_CNF(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3, int bit) const;
	int calc_CNF(const Indexes_vector &scope_Indexes, int bit) const;

	sSpecifiedIdentifier translate_CNFVariable(int cnf_variable) const;

	virtual void to_Screen(const sString &indent = "") const;

	static int calc_Log2(int value);
	static int calc_Exp2(int value);

    private:
	int m_N_States;
	int m_log_N_States;
	sIndexableIdentifier m_indexable_identifier;
    };


/*----------------------------------------------------------------------------*/

    inline int sIndexableStateIdentifier::get_Arity(void) const
    {
	return m_indexable_identifier.get_Arity();
    }


    inline int sIndexableStateIdentifier::get_First_CNFVariable(void) const
    {
	return m_indexable_identifier.get_First_CNFVariable();
    }


    inline int sIndexableStateIdentifier::get_CNFVariable_Count(void) const
    {
	return m_indexable_identifier.get_CNFVariable_Count();
    }

    
    inline int sIndexableStateIdentifier::get_StateCount(void) const
    {
	return m_N_States;
    }


   inline int sIndexableStateIdentifier::get_Log2_StateCount(void) const
    {
	return m_log_N_States;
    }


    inline const sIndexableIdentifier& sIndexableStateIdentifier::get_IndexableIdentifier(void) const
    {
	return m_indexable_identifier;
    }


/*----------------------------------------------------------------------------*/
// sStateClauseGenerator

    class sStateClauseGenerator
    {
    public:
	typedef std::vector<sSpecifiedStateIdentifier> SpecifiedStateIdentifiers_vector;
	typedef std::vector<int> States_vector;
	typedef std::map<int, sIndexableIdentifier, std::less<int> > AuxiliaryIdentifiers_map;

    public:
	sStateClauseGenerator(sVariableStore_CNF *variable_store);
	sStateClauseGenerator(const sStateClauseGenerator &clause_generator);
	const sStateClauseGenerator& operator=(const sStateClauseGenerator &clause_generator);

	virtual int count_Alignment(int &aux_Variable_cnt, const sSpecifiedStateIdentifier &spec_identifier) const;
	virtual int generate_Alignment(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier, bool string = false);

	virtual int count_Equality(int &aux_Variable_cnt, const sSpecifiedStateIdentifier &spec_identifier, int state) const;
	virtual int generate_Equality(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier, int state, bool string = false);

	virtual int count_Disequality(int &aux_Variable_cnt, const sSpecifiedStateIdentifier &spec_identifier, int state) const;
	virtual int generate_Disequality(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier, int state, bool string = false);

	virtual int count_Equality(int &aux_Variable_cnt, const sSpecifiedStateIdentifier &spec_identifier_A, const sSpecifiedStateIdentifier &spec_identifier_B) const;
	virtual int generate_Equality(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier_A, const sSpecifiedStateIdentifier &spec_identifier_B, bool string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
					      int                              state_THEN_B) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
						 int                              state_THEN_B,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2) const;

	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      int                              state_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      int                              state_THEN_B_2) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 int                              state_THEN_B_2,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      int                              state_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2,
						 bool                             string = false);
	
	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      int                              state_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      int                              state_THEN_B_2,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_3) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 int                              state_THEN_B_2,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_3,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      int                              state_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      int                              state_THEN_B_2,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
					      int                              state_THEN_B_3,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_4) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 int                              state_THEN_B_2,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
						 int                              state_THEN_B_3,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_4,
						 bool                             string = false);

	virtual int count_DifferenceConstraint(int                              &aux_Variable_cnt,
					       const sSpecifiedStateIdentifier  &spec_identifier_A,
					       SpecifiedStateIdentifiers_vector &spec_Identifiers_B) const;

	virtual int generate_DifferenceConstraint(FILE                             *fw,
						  const sSpecifiedStateIdentifier  &spec_identifier_A,
						  SpecifiedStateIdentifiers_vector &spec_Identifiers_B,
						  bool                              string = false);

	virtual int count_AllDifferenceConstraint(int                              &aux_Variable_cnt,
						  SpecifiedStateIdentifiers_vector &spec_Identifiers) const;

	virtual int generate_AllDifferenceConstraint(FILE                             *fw,
						     SpecifiedStateIdentifiers_vector &spec_Identifiers,
						     bool                              string = false);

	virtual int count_CaseSplitting(int                              &aux_Variable_cnt,
					SpecifiedStateIdentifiers_vector &split_Identifiers,
					States_vector                    &split_States) const;

	virtual int generate_CaseSplitting(FILE                             *fw,
					   SpecifiedStateIdentifiers_vector &split_Identifiers,
					   States_vector                    &split_States,
					   bool                              string = false);

	virtual int count_LargeConditionalEquality(int                              &aux_Variable_cnt,
						   const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						   int                               state_IF_B,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B);

	virtual int generate_LargeConditionalEquality(FILE                              *fw,
						      const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						      int                               state_IF_B,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B,
						      bool                              string = false);

	virtual int count_LargeConditionalEquality(int                              &aux_Variable_cnt,
						   const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						   int                               state_IF_B,
						   const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_1,
						   int                               state_THEN_B_1,
						   const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_2,
						   int                               state_THEN_B_2,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A_3,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B_3);

	virtual int generate_LargeConditionalEquality(FILE                              *fw,
						      const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						      int                               state_IF_B,
						      const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_1,
						      int                               state_THEN_B_1,
						      const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_2,
						      int                               state_THEN_B_2,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A_3,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B_3,
						      bool                              string = false);

	virtual int count_LEXLess_Constraint(int                             &aux_Variable_cnt,
					     const sSpecifiedStateIdentifier &spec_identifier_A,
					     const sSpecifiedStateIdentifier &spec_identifier_B) const;
	
	virtual int generate_LEXLess_Constraint(FILE                            *fw,
						const sSpecifiedStateIdentifier &spec_identifier_A,
						const sSpecifiedStateIdentifier &spec_identifier_B,
						bool                             string = false);

	sSpecifiedIdentifier translate_AuxiliaryCNFVariable(int cnf_variable) const;

	static int calc_Bit(int bit, int state);

    protected:
	sVariableStore_CNF *m_variable_store;

	int m_aux_Identifier_cnt;
	AuxiliaryIdentifiers_map m_auxiliary_Identifiers;
    };


/*----------------------------------------------------------------------------*/
// sAdvancedClauseGenerator

    class sAdvancedClauseGenerator
    : public sStateClauseGenerator
    {

    public:
	sAdvancedClauseGenerator(sVariableStore_CNF *variable_store);
	sAdvancedClauseGenerator(const sAdvancedClauseGenerator &clause_generator);
	const sAdvancedClauseGenerator& operator=(const sAdvancedClauseGenerator &clause_generator);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
					      int                              state_THEN_B) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      int                              state_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2,
						 bool                             string = false);

	virtual int count_ConditionalEquality(int                             &aux_Variable_cnt,
					      const sSpecifiedStateIdentifier &spec_identifier_IF_A,
					      int                              state_IF_B,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
					      int                              state_THEN_B_1,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
					      int                              state_THEN_B_2,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
					      int                              state_THEN_B_3,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
					      const sSpecifiedStateIdentifier &spec_identifier_THEN_B_4) const;
	
	virtual int generate_ConditionalEquality(FILE                            *fw,
						 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
						 int                              state_IF_B,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
						 int                              state_THEN_B_1,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
						 int                              state_THEN_B_2,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
						 int                              state_THEN_B_3,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
						 const sSpecifiedStateIdentifier &spec_identifier_THEN_B_4,
						 bool                             string = false);

	virtual int count_LargeConditionalEquality(int                              &aux_Variable_cnt,
						   const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						   int                               state_IF_B,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B);

	virtual int generate_LargeConditionalEquality(FILE                              *fw,
						      const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						      int                               state_IF_B,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B,
						      bool                              string = false);

	virtual int count_LargeConditionalEquality(int                              &aux_Variable_cnt,
						   const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						   int                               state_IF_B,
						   const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_1,
						   int                               state_THEN_B_1,
						   const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_2,
						   int                               state_THEN_B_2,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A_3,
						   SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B_3);

	virtual int generate_LargeConditionalEquality(FILE                              *fw,
						      const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
						      int                               state_IF_B,
						      const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_1,
						      int                               state_THEN_B_1,
						      const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_2,
						      int                               state_THEN_B_2,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A_3,
						      SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B_3,
						      bool                              string = false);

	virtual int count_AllDifferenceConstraint(int                              &aux_Variable_cnt,
						  SpecifiedStateIdentifiers_vector &spec_Identifiers) const;

	virtual int generate_AllDifferenceConstraint(FILE                             *fw,
						     SpecifiedStateIdentifiers_vector &spec_Identifiers,
						     bool                              string = false);
    };


/*----------------------------------------------------------------------------*/

} // namespace sReloc

#endif /* __CNF_H__ */
