/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                             reLOC 0.06-hanshin                             */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// multirobot.cpp / 0.06-hanshin_029
/*----------------------------------------------------------------------------*/
//
// Multirobot coordinated path-finding solving package.
//
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "config.h"
#include "compile.h"
#include "version.h"
#include "defs.h"
#include "types.h"
#include "result.h"
#include "cnf.h"
#include "multirobot.h"
#include "compress.h"
#include "statistics.h"

using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/
// sRobotArrangement

    sRobotArrangement::sRobotArrangement()
    {
	// nothing
    }
    

    sRobotArrangement::sRobotArrangement(int N_Vertices, int N_Robots, bool random)
	: m_robot_Locs(N_Robots + 1),
	  m_vertex_Occups(N_Vertices)
    {
	for (int i = 0; i < N_Vertices; ++i)
	{
	    m_vertex_Occups[i] = VACANT_VERTEX;
	}
	if (random)
	{
	    Vertices_vector Vertices;
	    for (int i = 0; i < N_Vertices; ++i)
	    {
		Vertices.push_back(i);
	    }

	    int remain = N_Vertices;
	    for (int r = N_Robots; r > 1;)
	    {
		int rnd = rand() % remain;
		place_Robot(--r, Vertices[rnd]);

		Vertices[rnd] = Vertices[--remain];
	    }
	}
	else
	{
	    for (int i = 0; i <= N_Robots; ++i)
	    {
		m_robot_Locs[i] = UNDEFINED_LOCATION;
	    }
	}
    }


    sRobotArrangement::sRobotArrangement(const sRobotArrangement &robot_arrangement)
	: m_robot_Locs(robot_arrangement.m_robot_Locs),
	  m_vertex_Occups(robot_arrangement.m_vertex_Occups)
    {
	// nothing
    }


    const sRobotArrangement& sRobotArrangement::operator=(const sRobotArrangement &robot_arrangement)
    {
	m_robot_Locs = robot_arrangement.m_robot_Locs;
	m_vertex_Occups = robot_arrangement.m_vertex_Occups;

	return *this;
    }


    bool sRobotArrangement::operator==(const sRobotArrangement &robot_arrangement) const
    {
	sASSERT(m_robot_Locs.size() == robot_arrangement.m_robot_Locs.size());

	Robots_vector::const_iterator robot_A, robot_B;
	
	for (robot_A = m_robot_Locs.begin(), robot_B = robot_arrangement.m_robot_Locs.begin();
	     robot_A != m_robot_Locs.end();
	     ++robot_A, ++robot_B)
	{
	    if (*robot_A != *robot_B)
	    {
		return false;
	    }
	}
	return true;
    }


    bool sRobotArrangement::operator<(const sRobotArrangement &robot_arrangement) const
    {
	sASSERT(m_robot_Locs.size() == robot_arrangement.m_robot_Locs.size());

	Robots_vector::const_iterator robot_A, robot_B;
	
	for (robot_A = m_robot_Locs.begin(), robot_B = robot_arrangement.m_robot_Locs.begin();
	     robot_A != m_robot_Locs.end();
	     ++robot_A, ++robot_B)
	{
	    if (*robot_A < *robot_B)
	    {
		return true;
	    }
	    else
	    {
		if (*robot_A > *robot_B)
		{
		    return false;
		}
	    }
	}

	return false;
    }


    int sRobotArrangement::get_RobotCount(void) const
    {
	return (m_robot_Locs.size() - 1);
    }

    
    int sRobotArrangement::get_VertexCount(void) const
    {
	return m_vertex_Occups.size();
    }
	

    int sRobotArrangement::get_RobotLocation(int robot_id) const
    {
	sASSERT(robot_id > 0 && robot_id < m_robot_Locs.size());

	return m_robot_Locs[robot_id];
    }


    int sRobotArrangement::get_VertexOccupancy(int vertex_id) const
    {
	sASSERT(vertex_id < m_vertex_Occups.size());

	return m_vertex_Occups[vertex_id];
    }


    void sRobotArrangement::place_Robot(int robot_id, int vertex_id)
    {
	sASSERT(robot_id > 0 && robot_id < m_robot_Locs.size());
	sASSERT(vertex_id < m_vertex_Occups.size());
	sASSERT(m_vertex_Occups[vertex_id] == VACANT_VERTEX);
	
	m_robot_Locs[robot_id] = vertex_id;
	m_vertex_Occups[vertex_id] = robot_id;
    }


    void sRobotArrangement::remove_Robot(int robot_id)
    {
	sASSERT(robot_id > 0 && robot_id < m_robot_Locs.size());

	m_vertex_Occups[m_robot_Locs[robot_id]] = VACANT_VERTEX;
	m_robot_Locs[robot_id] = UNDEFINED_LOCATION;
    }


    void sRobotArrangement::clean_Vertex(int vertex_id)
    {
	sASSERT(vertex_id < m_vertex_Occups.size());

	m_robot_Locs[m_vertex_Occups[vertex_id]] = UNDEFINED_LOCATION;
	m_vertex_Occups[vertex_id] = VACANT_VERTEX;
    }


    void sRobotArrangement::move_Robot(int robot_id, int dest_vertex_id)
    {
	sASSERT(robot_id > 0 && robot_id < m_robot_Locs.size());
	sASSERT(dest_vertex_id < m_vertex_Occups.size());
	#ifdef sDEBUG
	{
	    if (m_vertex_Occups[dest_vertex_id] != VACANT_VERTEX)
	    {
		printf("%d # %d -> %d\n", robot_id, m_robot_Locs[robot_id], dest_vertex_id);
	    }
	}
        #endif
	sASSERT(m_vertex_Occups[dest_vertex_id] == VACANT_VERTEX);

	m_vertex_Occups[m_robot_Locs[robot_id]] = VACANT_VERTEX;
	m_robot_Locs[robot_id] = dest_vertex_id;
	m_vertex_Occups[dest_vertex_id] = robot_id;

        #ifdef sSTATISTICS
	{
	    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_move_Executions;
	}
	#endif
    }


    bool sRobotArrangement::verify_Move(int robot_id, int dest_vertex_id, const sUndirectedGraph &graph) const
    {
	if (   (robot_id > 0 && robot_id < m_robot_Locs.size())
	    && (dest_vertex_id < m_vertex_Occups.size())
	    && (m_vertex_Occups[dest_vertex_id] == VACANT_VERTEX)
	    && (m_vertex_Occups[m_robot_Locs[robot_id]] == robot_id)
	    && graph.is_Adjacent(m_robot_Locs[robot_id], dest_vertex_id))
	{
	    return true;
	}

	#ifdef sDEBUG
	{
	    printf("%d # %d -> %d (adjacent:%d)\n", robot_id, m_robot_Locs[robot_id], dest_vertex_id, graph.is_Adjacent(m_robot_Locs[robot_id], dest_vertex_id) ? 1 : 0);
	}
	#endif

	return false;
    }


    void sRobotArrangement::to_Screen(const sString &indent) const
    {
	to_Stream(stdout, indent);
    }


    void sRobotArrangement::to_Stream(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%sRobot arrangement: (|R| = %ld, |V| = %ld) [\n", indent.c_str(), m_robot_Locs.size() - 1, m_vertex_Occups.size());
	fprintf(fw, "%s%s robot locations: {", indent.c_str(), sRELOC_INDENT.c_str());
	
	int N_Robots_1 = m_robot_Locs.size();
	for (int i = 1; i < N_Robots_1; ++i)
	{
	    fprintf(fw, "%d#%d ", i, m_robot_Locs[i]);
	}
	fprintf(fw, "}\n");

	fprintf(fw, "%s%s vertex occupancy: {", indent.c_str(), sRELOC_INDENT.c_str());
	
	int N_Vertices = m_vertex_Occups.size();
	for (int i = 0; i < N_Vertices; ++i)
	{
	    fprintf(fw, "%d#%d ", m_vertex_Occups[i], i);
	}
	fprintf(fw, "}\n");

	fprintf(fw, "%s]\n", indent.c_str());
    }


    sResult sRobotArrangement::to_File_multirobot(const sString &filename, const sString &indent) const
    {
	FILE *fw;

	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sROBOT_ARRANGEMENT_OPEN_ERROR;
	}
	
	to_Stream_multirobot(fw, indent);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    void sRobotArrangement::to_Stream_multirobot(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%sV =\n", indent.c_str());
	
	int N_Vertices = m_vertex_Occups.size();
	for (int i = 0; i < N_Vertices; ++i)
	{
	    fprintf(fw, "(%d:-1)[%d:-1:-1]\n", i, m_vertex_Occups[i]);
	}
    }


    sResult sRobotArrangement::from_File_multirobot(const sString &filename, int component)
    {
	sResult result;
	FILE *fr;

	if ((fr = fopen(filename.c_str(), "r")) == NULL)
	{
	    return sROBOT_ARRANGEMENT_OPEN_ERROR;
	}
	
	result = from_Stream_multirobot(fr, component);
	if (sFAILED(result))
	{
	    fclose(fr);
	    return result;
	}
	fclose(fr);

	return sRESULT_SUCCESS;
    }


    sResult sRobotArrangement::from_Stream_multirobot(FILE *fr, int component)
    {
	m_robot_Locs.clear();
	m_vertex_Occups.clear();

	int N_Robots = 0;
	int N_Vertices = 0;

	int c = fgetc(fr);

	while (c != 'V')
	{
	    if (c != '\n')
	    {
		while(fgetc(fr) != '\n');
	    }
	    c = fgetc(fr);
	}

	fscanf(fr, " =\n");

	long position = ftell(fr);
	c = fgetc(fr);

	while (c == '(')
	{
	    int vertex_id, cycle_id, robot_id;

	    switch (component)
	    {
	    case 0:
	    {
		fscanf(fr, "%d:%d)[%d", &vertex_id, &cycle_id, &robot_id);
		break;
	    }
	    case 1:
	    {
		int dummy_robot_1_id;
		fscanf(fr, "%d:%d)[%d:%d", &vertex_id, &cycle_id, &dummy_robot_1_id, &robot_id);
		break;
	    }
	    case 2:
	    {
		int dummy_robot_1_id, dummy_robot_2_id;
		fscanf(fr, "%d:%d)[%d:%d:%d", &vertex_id, &cycle_id, &dummy_robot_1_id, &dummy_robot_2_id, &robot_id);
		break;
	    }
	    default:
	    {
		sASSERT(false);
		break;
	    }
	    }

	    if (robot_id > 0)
	    {
		++N_Robots;
	    }
	    ++N_Vertices;
	    if (c != '\n')
	    {
		while(fgetc(fr) != '\n');
	    }
	    c = fgetc(fr);
	}

	if (fseek(fr, position, SEEK_SET) != 0)
	{
	    return sROBOT_ARRANGEMENT_SEEK_ERROR;
	}
	c = fgetc(fr);

	m_robot_Locs.resize(N_Robots + 1, UNDEFINED_LOCATION);
	m_vertex_Occups.resize(N_Vertices, VACANT_VERTEX);

	while (c == '(')
	{
	    int vertex_id, cycle_id, robot_id;

	    switch (component)
	    {
	    case 0:
	    {
		fscanf(fr, "%d:%d)[%d", &vertex_id, &cycle_id, &robot_id);
		break;
	    }
	    case 1:
	    {
		int dummy_robot_1_id;
		fscanf(fr, "%d:%d)[%d:%d", &vertex_id, &cycle_id, &dummy_robot_1_id, &robot_id);
		break;
	    }
	    case 2:
	    {
		int dummy_robot_1_id, dummy_robot_2_id;
		fscanf(fr, "%d:%d)[%d:%d:%d", &vertex_id, &cycle_id, &dummy_robot_1_id, &dummy_robot_2_id, &robot_id);
		break;
	    }
	    default:
	    {
		sASSERT(false);
		break;
	    }
	    }

	    if (robot_id > 0)
	    {
		m_robot_Locs[robot_id] = vertex_id;
		m_vertex_Occups[vertex_id] = robot_id;
	    }
	    if (c != '\n')
	    {
		while(fgetc(fr) != '\n');
	    }
	    c = fgetc(fr);
	}

	return sRESULT_SUCCESS;
    }


/*----------------------------------------------------------------------------*/
// sMultirobotEncodingContext_CNFsat

    const int sMultirobotEncodingContext_CNFsat::UNDEFINED_LAYER_COUNT = -1;


/*----------------------------------------------------------------------------*/

    sMultirobotEncodingContext_CNFsat::sMultirobotEncodingContext_CNFsat()
	: m_N_Layers(UNDEFINED_LAYER_COUNT)
	, m_state_clause_generator(NULL)
	, m_advanced_clause_generator(NULL)
    {
	m_state_Identifiers[sINT_32_MAX] = NULL;
	switchTo_StandardGeneratingMode();
    }


    sMultirobotEncodingContext_CNFsat::sMultirobotEncodingContext_CNFsat(int N_Layers)
	: m_N_Layers(N_Layers)
	, m_state_clause_generator(&m_variable_store)
	, m_advanced_clause_generator(&m_variable_store)
    {
	m_state_Identifiers[sINT_32_MAX] = NULL;
	switchTo_StandardGeneratingMode();
    }


    sMultirobotEncodingContext_CNFsat::sMultirobotEncodingContext_CNFsat(const sMultirobotEncodingContext_CNFsat &encoding_context)
	: m_N_Layers(encoding_context.m_N_Layers)
	, m_variable_store(encoding_context.m_variable_store)
	, m_state_clause_generator(encoding_context.m_state_clause_generator)
	, m_advanced_clause_generator(encoding_context.m_advanced_clause_generator)
	, m_vertex_occupancy(encoding_context.m_vertex_occupancy)
	, m_robot_location(encoding_context.m_robot_location)
	, m_transition_action(encoding_context.m_transition_action)
	, m_transition_Actions(encoding_context.m_transition_Actions)
    {
	register_InternalIdentifiers();
	switch_GeneratingMode(encoding_context.get_GeneratingMode());
    }


    const sMultirobotEncodingContext_CNFsat& sMultirobotEncodingContext_CNFsat::operator=(const sMultirobotEncodingContext_CNFsat &encoding_context)
    {
	m_N_Layers = encoding_context.m_N_Layers;
	m_variable_store = encoding_context.m_variable_store;
	m_clause_generator = encoding_context.m_clause_generator;
	m_state_clause_generator = encoding_context.m_state_clause_generator;
	m_advanced_clause_generator = encoding_context.m_advanced_clause_generator;
	m_vertex_occupancy = encoding_context.m_vertex_occupancy;
	m_robot_location = encoding_context.m_robot_location;
	m_transition_action = encoding_context.m_transition_action;
	m_transition_Actions = encoding_context.m_transition_Actions;
	switch_GeneratingMode(encoding_context.get_GeneratingMode());

	register_InternalIdentifiers();

	return *this;
    }


    sMultirobotEncodingContext_CNFsat::GeneratingMode sMultirobotEncodingContext_CNFsat::get_GeneratingMode(void) const
    {
	return m_generating_mode;
    }

	
    void sMultirobotEncodingContext_CNFsat::switch_GeneratingMode(GeneratingMode generating_mode)
    {
	m_generating_mode = generating_mode;

	switch (generating_mode)
	{
	case GENERATING_STANDARD:
	{
	    m_clause_generator = &m_state_clause_generator;
	    break;
	}
	case GENERATING_ADVANCED:
	{
	    m_clause_generator = &m_advanced_clause_generator;
	    break;
	}
	default:
	{
	    sASSERT(false);
	    break;
	}
	}
    }


    void sMultirobotEncodingContext_CNFsat::switchTo_StandardGeneratingMode(void)
    {
	m_clause_generator = &m_state_clause_generator;
	m_generating_mode = GENERATING_STANDARD;
    }


    void sMultirobotEncodingContext_CNFsat::switchTo_AdvancedGeneratingMode(void)
    {
	m_clause_generator = &m_advanced_clause_generator;
	m_generating_mode = GENERATING_ADVANCED;
    }


    void sMultirobotEncodingContext_CNFsat::register_InternalIdentifiers(void)
    {
	m_state_Identifiers.clear();
	m_state_Identifiers[sINT_32_MAX] = NULL;

	register_TranslateIdentifier(m_vertex_occupancy);
	register_TranslateIdentifier(m_robot_location);

	for (StateIdentifiers_vector::iterator transition_action = m_transition_Actions.begin(); transition_action != m_transition_Actions.end(); ++transition_action)
	{
	    register_TranslateIdentifier(*transition_action);
	}
    }


    void sMultirobotEncodingContext_CNFsat::register_TranslateIdentifier(sIndexableStateIdentifier &state_identifier)
    {
	if (state_identifier.get_First_CNFVariable() >= 1)
	{
	    m_state_Identifiers[state_identifier.get_First_CNFVariable()] = &state_identifier;
	}
    }


    sSpecifiedIdentifier sMultirobotEncodingContext_CNFsat::translate_CNF_Variable(int cnf_variable) const
    {
	sSpecifiedIdentifier state_translation;
	StateIdentifiers_map::const_iterator state_identifier = m_state_Identifiers.upper_bound(cnf_variable);

	/*
	printf("cnf_var:%d,%ld\n", cnf_variable, m_state_Identifiers.size());
	for (StateIdentifiers_map::const_iterator identifier = m_state_Identifiers.begin(); identifier != m_state_Identifiers.end(); ++identifier)
	{
	    if (identifier->second != NULL)
	    {
		//		identifier->second->to_Screen();
		printf("%d\n", identifier->first);
	    }
	}
	*/
	
	if (state_identifier != m_state_Identifiers.end())
	{
	    if (state_identifier->first > cnf_variable)
	    {
		--state_identifier;
	    }
	    state_translation = state_identifier->second->translate_CNFVariable(cnf_variable);
	}
	if (!state_translation.is_Null())
	{
	    return state_translation;
	}
	else
	{
	    return m_clause_generator->translate_AuxiliaryCNFVariable(cnf_variable);
	}

	return sSpecifiedIdentifier();
    }


/*----------------------------------------------------------------------------*/
// sMultirobotInstance

    sMultirobotInstance::sMultirobotInstance()
    {
	// nothing
    }


    sMultirobotInstance::sMultirobotInstance(const sUndirectedGraph &environment, const sRobotArrangement &initial_arrangement, const sRobotArrangement &goal_arrangement)
	: m_environment(environment),
	  m_initial_arrangement(initial_arrangement),
	  m_goal_arrangement(goal_arrangement)
    {
	sASSERT(environment.get_VertexCount() == initial_arrangement.get_VertexCount() && environment.get_VertexCount() == goal_arrangement.get_VertexCount());
    }


    sMultirobotInstance::sMultirobotInstance(const sMultirobotInstance &multirobot_instance)
	: m_environment(multirobot_instance.m_environment),
	  m_initial_arrangement(multirobot_instance.m_initial_arrangement),
	  m_goal_arrangement(multirobot_instance.m_goal_arrangement)
    {
	// nothing
    }


    const sMultirobotInstance& sMultirobotInstance::operator=(const sMultirobotInstance &multirobot_instance)
    {
	m_environment = multirobot_instance.m_environment;
	m_initial_arrangement = multirobot_instance.m_initial_arrangement;
	m_goal_arrangement = multirobot_instance.m_goal_arrangement;

	return *this;
    }


    void sMultirobotInstance::to_Screen(const sString &indent) const
    {
	to_Stream(stdout, indent);
    }


    void sMultirobotInstance::to_Screen_multirobot(const sString &indent) const
    {
	to_Stream_multirobot(stdout, indent);
    }

    void sMultirobotInstance::to_Screen_domainPDDL(const sString &indent) const
    {
	to_Stream_domainPDDL(stdout, indent);
    }


    void sMultirobotInstance::to_Screen_problemPDDL(const sString &indent) const
    {
	to_Stream_problemPDDL(stdout, indent);
    }


    void sMultirobotInstance::to_Screen_InverseCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_InverseCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_AdvancedCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_AdvancedCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_DifferentialCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_DifferentialCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_BijectionCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_BijectionCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_HeuristicDifferentialCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_HeuristicDifferentialCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_HeuristicBijectionCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_HeuristicBijectionCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_HeuristicAdvancedCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_HeuristicAdvancedCNFsat(stdout, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Screen_PuzzleCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	to_Stream_PuzzleCNFsat(stdout, encoding_context, indent, verbose);
    }


    sResult sMultirobotInstance::to_File(const sString &filename, const sString &indent) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_OPEN_ERROR;
	}
	to_Stream(fw, indent);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_multirobot(const sString &filename, const sString &indent) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_OPEN_ERROR;
	}
	to_Stream_multirobot(fw, indent);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_domainPDDL(const sString &filename, const sString &indent) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_PDDL_OPEN_ERROR;
	}
	to_Stream_domainPDDL(fw, indent);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_problemPDDL(const sString &filename, const sString &indent) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_PDDL_OPEN_ERROR;
	}
	to_Stream_problemPDDL(fw, indent);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_InverseCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_InverseCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_AdvancedCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_AdvancedCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_DifferentialCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_DifferentialCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_BijectionCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_BijectionCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_HeuristicDifferentialCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_HeuristicDifferentialCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_HeuristicBijectionCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_HeuristicBijectionCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_HeuristicAdvancedCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_HeuristicAdvancedCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotInstance::to_File_PuzzleCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	FILE *fw;
	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_CNF_OPEN_ERROR;
	}
	to_Stream_PuzzleCNFsat(fw, encoding_context, indent, verbose);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    void sMultirobotInstance::to_Stream(FILE *fw, const sString &indent) const
    {       
	fprintf(fw, "%sMultirobot instance: [\n", indent.c_str());
	fprintf(fw, "%s%sEnvironment:\n", indent.c_str(), sRELOC_INDENT.c_str());
	m_environment.to_Stream_vertices(fw, indent + sRELOC_INDENT + sRELOC_INDENT);
	fprintf(fw, "%s%sInitial arrangement:\n", indent.c_str(), sRELOC_INDENT.c_str());
	m_initial_arrangement.to_Stream(fw, indent + sRELOC_INDENT + sRELOC_INDENT);
	fprintf(fw, "%s%sGoal arrangement:\n", indent.c_str(), sRELOC_INDENT.c_str());
	m_goal_arrangement.to_Stream(fw, indent + sRELOC_INDENT + sRELOC_INDENT);
	fprintf(fw, "%s]\n", indent.c_str());
    }


    void sMultirobotInstance::to_Stream_multirobot(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%sV =\n", indent.c_str());
	
	int N_Vertices = m_initial_arrangement.m_vertex_Occups.size();
	for (int i = 0; i < N_Vertices; ++i)
	{
	    fprintf(fw, "(%d:-1)[%d:%d:%d]\n", i,
		                               m_initial_arrangement.m_vertex_Occups[i],
		                               m_goal_arrangement.m_vertex_Occups[i],
		                               m_goal_arrangement.m_vertex_Occups[i]);
	}
	m_environment.to_Stream_multirobot(fw, indent);
    }


    void sMultirobotInstance::to_Stream_domainPDDL(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%s(define (domain multirobot)\n", indent.c_str());
	fprintf(fw, "%s%s(:predicates\n", indent.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s(adjacent ?u ?v)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s(robot_location ?r ?v)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s(no_robot ?v)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s)\n", indent.c_str(), sRELOC_INDENT.c_str());

	fprintf(fw, "%s%s(:action move\n", indent.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s:parameters (?r ?u ?v)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s:precondition (and\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s%s(robot_location ?r ?u) (no_robot ?v) (adjacent ?u ?v)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s:effect (and\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s%s(robot_location ?r ?v) (no_robot ?u) (not (no_robot ?v))\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s%s)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str());
	fprintf(fw, "%s%s)\n", indent.c_str(), sRELOC_INDENT.c_str());

	fprintf(fw, "%s)\n", indent.c_str());
    }


    void sMultirobotInstance::to_Stream_problemPDDL(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%s(define (problem multirobot_instance)\n", indent.c_str());
	fprintf(fw, "%s%s(:domain multirobot)\n", indent.c_str(), sRELOC_INDENT.c_str());

	fprintf(fw, "%s%s(:objects\n", indent.c_str(), sRELOC_INDENT.c_str());

	int N_Vertices = m_environment.get_VertexCount();
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    fprintf(fw, "%s%s%sv%d\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), vertex_id);
	}

	int N_Robots = m_initial_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    fprintf(fw, "%s%s%sr%d\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), robot_id);
	}
	fprintf(fw, "%s%s)\n", indent.c_str(), sRELOC_INDENT.c_str());

	fprintf(fw, "%s%s(:init\n", indent.c_str(), sRELOC_INDENT.c_str());

	for (sUndirectedGraph::Edges_list::const_iterator edge = m_environment.m_Edges.begin(); edge != m_environment.m_Edges.end(); ++edge)
	{
	    fprintf(fw, "%s%s%s(adjacent v%d v%d)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), edge->m_arc_vu.m_target->m_id, edge->m_arc_uv.m_target->m_id);
	    fprintf(fw, "%s%s%s(adjacent v%d v%d)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), edge->m_arc_uv.m_target->m_id, edge->m_arc_vu.m_target->m_id);
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    fprintf(fw, "%s%s%s(robot_location r%d v%d)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), robot_id, m_initial_arrangement.get_RobotLocation(robot_id));
	}
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);

	    if (robot_id == sRobotArrangement::VACANT_VERTEX)
	    {
		fprintf(fw, "%s%s%s(no_robot v%d)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), vertex_id);
	    }
	}
	fprintf(fw, "%s%s)\n", indent.c_str(), sRELOC_INDENT.c_str());

	fprintf(fw, "%s%s(:goal (and \n", indent.c_str(), sRELOC_INDENT.c_str());

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    fprintf(fw, "%s%s%s(robot_location r%d v%d)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), robot_id, m_goal_arrangement.get_RobotLocation(robot_id));
	}
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);

	    if (robot_id == sRobotArrangement::VACANT_VERTEX)
	    {
		fprintf(fw, "%s%s%s(no_robot v%d)\n", indent.c_str(), sRELOC_INDENT.c_str(), sRELOC_INDENT.c_str(), vertex_id);
	    }
	}
	fprintf(fw, "%s%s))\n", indent.c_str(), sRELOC_INDENT.c_str());

	fprintf(fw, "%s)\n", indent.c_str());
    }


    void sMultirobotInstance::to_Stream_InverseCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &sUNUSED(indent), bool verbose) const
    {
	int Clause_cnt = 0;
	int aux_Variable_cnt = 0;
 
	int N_Vertices = m_environment.get_VertexCount();
	int N_Robots = m_initial_arrangement.get_RobotCount();

	sIndexableStateIdentifier vertex_occupancy(&encoding_context.m_variable_store, "vertex_occupancy", N_Robots + 1, sIntegerScope(0, N_Vertices), sIntegerScope(0, encoding_context.m_N_Layers - 1));
	encoding_context.m_vertex_occupancy = vertex_occupancy;
	encoding_context.register_TranslateIdentifier(encoding_context.m_vertex_occupancy);

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
	    }
	}

	encoding_context.m_transition_Actions.resize(N_Vertices);
	
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    sIndexableStateIdentifier transition_action(&encoding_context.m_variable_store, "transition_action-" + sInt_32_to_String(vertex_id), 2 * m_environment.get_Vertex(vertex_id)->calc_NeighborCount() + 1, sIntegerScope(0, encoding_context.m_N_Layers - 1)); 
	    encoding_context.m_transition_Actions[vertex_id] = transition_action;
	    encoding_context.register_TranslateIdentifier(encoding_context.m_transition_Actions[vertex_id]);

	    for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	    {		    
		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&transition_action, sIntegerIndex(layer)));
	    }
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{	
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {
		int N_vertex_Neighbors = m_environment.get_Vertex(vertex_id)->calc_NeighborCount();
		const sVertex::Neighbors_list &out_Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;

		int out_neighbor_order = 0;

		for (sVertex::Neighbors_list::const_iterator out_neighbor = out_Neighbors.begin(); out_neighbor != out_Neighbors.end(); ++out_neighbor)
		{
		    int neighbor_id = (*out_neighbor)->m_target->m_id;
		    int N_neighbor_Neighbors = m_environment.get_Vertex(neighbor_id)->calc_NeighborCount();
		    int in_neighbor_order = m_environment.get_Vertex(neighbor_id)->calc_NeighborOrder(vertex_id);

		    Clause_cnt += encoding_context.m_clause_generator->count_ConditionalEquality(aux_Variable_cnt,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
												 out_neighbor_order,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
												 N_neighbor_Neighbors + in_neighbor_order,
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer)),
												 0,
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
												 0,
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer + 1)),
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
		    
		    Clause_cnt += encoding_context.m_clause_generator->count_ConditionalEquality(aux_Variable_cnt,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
												 N_vertex_Neighbors + out_neighbor_order,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
												 in_neighbor_order);

		    ++out_neighbor_order;
		}
		
		Clause_cnt += encoding_context.m_clause_generator->count_ConditionalEquality(aux_Variable_cnt,
											     sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
											     2 * N_vertex_Neighbors,
											     sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
											     sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
	    }
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(0)), robot_id);
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), robot_id);
	}

	int N_cnf_Variables = encoding_context.m_variable_store.get_Last_CNFVariable() + aux_Variable_cnt - 1;

	fprintf(fw, "c %s : multirobot inverse SAT encoding\n", sPRODUCT);
	fprintf(fw, "c %s\n", sCOPYRIGHT);
	fprintf(fw, "c number of layers = %d\n", encoding_context.m_N_Layers);
	fprintf(fw, "c number of clauses / number of variables = %.3f\n", (double)Clause_cnt / N_cnf_Variables);
	fprintf(fw, "c number of visible variables = %d\n", N_cnf_Variables - aux_Variable_cnt);
	fprintf(fw, "c number of hidden variables = %d\n", aux_Variable_cnt);
	to_Stream(fw, "c ");
	fprintf(fw, "p cnf %d %d\n", N_cnf_Variables, Clause_cnt);

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)), verbose);
	    }
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	    {		    
		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)), verbose);
	    }
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{	
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {
		int N_vertex_Neighbors = m_environment.get_Vertex(vertex_id)->calc_NeighborCount();
		const sVertex::Neighbors_list &out_Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;

		int out_neighbor_order = 0;

		for (sVertex::Neighbors_list::const_iterator out_neighbor = out_Neighbors.begin(); out_neighbor != out_Neighbors.end(); ++out_neighbor)
		{
		    int neighbor_id = (*out_neighbor)->m_target->m_id;
		    int N_neighbor_Neighbors = m_environment.get_Vertex(neighbor_id)->calc_NeighborCount();
		    int in_neighbor_order = m_environment.get_Vertex(neighbor_id)->calc_NeighborOrder(vertex_id);

		    encoding_context.m_clause_generator->generate_ConditionalEquality(fw,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
										      out_neighbor_order,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
										      N_neighbor_Neighbors + in_neighbor_order,
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer)),
										      0,
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
										      0,
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer + 1)),
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)),
										      verbose);
		    
		    encoding_context.m_clause_generator->generate_ConditionalEquality(fw,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
										      N_vertex_Neighbors + out_neighbor_order,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
										      in_neighbor_order,
										      verbose);
		    
		    ++out_neighbor_order;
		}
		encoding_context.m_clause_generator->generate_ConditionalEquality(fw,
										  sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
										  2 * N_vertex_Neighbors,
										  sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
										  sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)),
										  verbose);
	    }
	}
	
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(0)), robot_id, verbose);
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), robot_id, verbose);
	} 
    }


    void sMultirobotInstance::to_Stream_AdvancedCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	encoding_context.switchTo_AdvancedGeneratingMode();
	to_Stream_InverseCNFsat(fw, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Stream_DifferentialCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &sUNUSED(indent), bool verbose) const
    {
	int Clause_cnt = 0;
	int aux_Variable_cnt = 0;

	int N_Vertices = m_environment.get_VertexCount();
	int N_Robots = m_initial_arrangement.get_RobotCount();

	sIndexableStateIdentifier robot_location(&encoding_context.m_variable_store, "robot_location", N_Vertices, sIntegerScope(1, N_Robots), sIntegerScope(0, encoding_context.m_N_Layers - 1));
	encoding_context.m_robot_location = robot_location;
	encoding_context.register_TranslateIdentifier(encoding_context.m_robot_location);

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
	    }
	}
 
	for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	{
	    sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_diff_Identifiers;
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
	    }		
	    Clause_cnt += encoding_context.m_clause_generator->count_AllDifferenceConstraint(aux_Variable_cnt, robot_diff_Identifiers);
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector prev_robot_diff_Identifiers;
		int prev_robot_id;
		for (prev_robot_id = 1; prev_robot_id < robot_id; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		for (++prev_robot_id; prev_robot_id <= N_Robots; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		Clause_cnt += encoding_context.m_clause_generator->count_DifferenceConstraint(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)), prev_robot_diff_Identifiers);

		for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
		{
		    sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_case_split_Identifiers;
		    sStateClauseGenerator::States_vector robot_case_split_States;

		    robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
		    robot_case_split_States.push_back(vertex_id);

		    robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
		    robot_case_split_States.push_back(vertex_id);

		    const sVertex::Neighbors_list &Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;
		    
		    for (sVertex::Neighbors_list::const_iterator neighbor = Neighbors.begin(); neighbor != Neighbors.end(); ++neighbor)
		    {
			int neighbor_id = (*neighbor)->m_target->m_id;

			robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
			robot_case_split_States.push_back(neighbor_id);
		    }
		    Clause_cnt += encoding_context.m_clause_generator->count_CaseSplitting(aux_Variable_cnt, robot_case_split_Identifiers, robot_case_split_States);
		}
	    }
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_initial_arrangement.get_RobotLocation(robot_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(0)), vertex_id);
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_goal_arrangement.get_RobotLocation(robot_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), vertex_id);
	}
	
	int N_cnf_Variables = encoding_context.m_variable_store.get_Last_CNFVariable() + aux_Variable_cnt - 1;

	fprintf(fw, "c %s : multirobot differential SAT encoding\n", sPRODUCT);
	fprintf(fw, "c %s\n", sCOPYRIGHT);
	fprintf(fw, "c number of layers = %d\n", encoding_context.m_N_Layers);
	fprintf(fw, "c number of clauses / number of variables = %.3f\n", (double)Clause_cnt / N_cnf_Variables);
	fprintf(fw, "c number of visible variables = %d\n", N_cnf_Variables - aux_Variable_cnt);
	fprintf(fw, "c number of hidden variables = %d\n", aux_Variable_cnt);
	to_Stream(fw, "c ");
	fprintf(fw, "p cnf %d %d\n", N_cnf_Variables, Clause_cnt);

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)), verbose);
	    }
	}
 
	for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	{
	    sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_diff_Identifiers;
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
	    }		
	    encoding_context.m_clause_generator->generate_AllDifferenceConstraint(fw, robot_diff_Identifiers, verbose);
	}
	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector prev_robot_diff_Identifiers;
		int prev_robot_id;
		for (prev_robot_id = 1; prev_robot_id < robot_id; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		for (++prev_robot_id; prev_robot_id <= N_Robots; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		encoding_context.m_clause_generator->generate_DifferenceConstraint(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)), prev_robot_diff_Identifiers, verbose);

		for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
		{
		    sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_case_split_Identifiers;
		    sStateClauseGenerator::States_vector robot_case_split_States;

		    robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
		    robot_case_split_States.push_back(vertex_id);

		    robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
		    robot_case_split_States.push_back(vertex_id);

		    const sVertex::Neighbors_list &Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;
		    
		    for (sVertex::Neighbors_list::const_iterator neighbor = Neighbors.begin(); neighbor != Neighbors.end(); ++neighbor)
		    {
			int neighbor_id = (*neighbor)->m_target->m_id;

			robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
			robot_case_split_States.push_back(neighbor_id);
		    }
		    encoding_context.m_clause_generator->generate_CaseSplitting(fw, robot_case_split_Identifiers, robot_case_split_States, verbose);
		}
	    }
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_initial_arrangement.get_RobotLocation(robot_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(0)), vertex_id, verbose);
	}
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_goal_arrangement.get_RobotLocation(robot_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), vertex_id, verbose);
	}
    }


    void sMultirobotInstance::to_Stream_BijectionCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	encoding_context.switchTo_AdvancedGeneratingMode();
	to_Stream_DifferentialCNFsat(fw, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Stream_HeuristicDifferentialCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &sUNUSED(indent), bool verbose) const
    {
	int Clause_cnt = 0;
	int aux_Variable_cnt = 0;

	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	m_environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	int N_Vertices = m_environment.get_VertexCount();
	int N_Robots = m_initial_arrangement.get_RobotCount();

	sIndexableStateIdentifier robot_location(&encoding_context.m_variable_store, "robot_location", N_Vertices, sIntegerScope(1, N_Robots), sIntegerScope(0, encoding_context.m_N_Layers - 1));
	encoding_context.m_robot_location = robot_location;
	encoding_context.register_TranslateIdentifier(encoding_context.m_robot_location);

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
	    }
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	{
	    sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_diff_Identifiers;
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
	    }		
	    Clause_cnt += encoding_context.m_clause_generator->count_AllDifferenceConstraint(aux_Variable_cnt, robot_diff_Identifiers);
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector prev_robot_diff_Identifiers;
		int prev_robot_id;
		for (prev_robot_id = 1; prev_robot_id < robot_id; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		for (++prev_robot_id; prev_robot_id <= N_Robots; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		Clause_cnt += encoding_context.m_clause_generator->count_DifferenceConstraint(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)), prev_robot_diff_Identifiers);

		for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
		{
		    if (all_pairs_Distances[m_initial_arrangement.get_RobotLocation(robot_id)][vertex_id] <= layer && all_pairs_Distances[vertex_id][m_goal_arrangement.get_RobotLocation(robot_id)] <= encoding_context.m_N_Layers - layer)
		    {
			sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_case_split_Identifiers;
			sStateClauseGenerator::States_vector robot_case_split_States;
			
			robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
			robot_case_split_States.push_back(vertex_id);
			
			robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
			robot_case_split_States.push_back(vertex_id);
			
			const sVertex::Neighbors_list &Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;
			
			for (sVertex::Neighbors_list::const_iterator neighbor = Neighbors.begin(); neighbor != Neighbors.end(); ++neighbor)
			{
			    int neighbor_id = (*neighbor)->m_target->m_id;
			    
			    robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
			    robot_case_split_States.push_back(neighbor_id);
			}
			Clause_cnt += encoding_context.m_clause_generator->count_CaseSplitting(aux_Variable_cnt, robot_case_split_Identifiers, robot_case_split_States);
		    }
		    else
		    {
			Clause_cnt += encoding_context.m_clause_generator->count_Disequality(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)), vertex_id);
		    }
		}
	    }
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_initial_arrangement.get_RobotLocation(robot_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(0)), vertex_id);
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_goal_arrangement.get_RobotLocation(robot_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), vertex_id);
	}
	
	int N_cnf_Variables = encoding_context.m_variable_store.get_Last_CNFVariable() + aux_Variable_cnt - 1;

	fprintf(fw, "c %s : multirobot differential SAT encoding\n", sPRODUCT);
	fprintf(fw, "c %s\n", sCOPYRIGHT);
	fprintf(fw, "c number of layers = %d\n", encoding_context.m_N_Layers);
	fprintf(fw, "c number of clauses / number of variables = %.3f\n", (double)Clause_cnt / N_cnf_Variables);
	fprintf(fw, "c number of visible variables = %d\n", N_cnf_Variables - aux_Variable_cnt);
	fprintf(fw, "c number of hidden variables = %d\n", aux_Variable_cnt);
	to_Stream(fw, "c ");
	fprintf(fw, "p cnf %d %d\n", N_cnf_Variables, Clause_cnt);

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)), verbose);
	    }
	}
 
	for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	{
	    sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_diff_Identifiers;
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
	    }		
	    encoding_context.m_clause_generator->generate_AllDifferenceConstraint(fw, robot_diff_Identifiers, verbose);
	}
	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector prev_robot_diff_Identifiers;
		int prev_robot_id;
		for (prev_robot_id = 1; prev_robot_id < robot_id; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		for (++prev_robot_id; prev_robot_id <= N_Robots; ++prev_robot_id)
		{
		    prev_robot_diff_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(prev_robot_id), sIntegerIndex(layer)));
		}
		encoding_context.m_clause_generator->generate_DifferenceConstraint(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)), prev_robot_diff_Identifiers, verbose);

		for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
		{
		    if (all_pairs_Distances[m_initial_arrangement.get_RobotLocation(robot_id)][vertex_id] <= layer && all_pairs_Distances[vertex_id][m_goal_arrangement.get_RobotLocation(robot_id)] <= encoding_context.m_N_Layers - layer)
		    {
			sStateClauseGenerator::SpecifiedStateIdentifiers_vector robot_case_split_Identifiers;
			sStateClauseGenerator::States_vector robot_case_split_States;
			
			robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)));
			robot_case_split_States.push_back(vertex_id);
			
			robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
			robot_case_split_States.push_back(vertex_id);
			
			const sVertex::Neighbors_list &Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;
			
			for (sVertex::Neighbors_list::const_iterator neighbor = Neighbors.begin(); neighbor != Neighbors.end(); ++neighbor)
			{
			    int neighbor_id = (*neighbor)->m_target->m_id;
			    
			    robot_case_split_Identifiers.push_back(sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer + 1)));
			    robot_case_split_States.push_back(neighbor_id);
			}
			encoding_context.m_clause_generator->generate_CaseSplitting(fw, robot_case_split_Identifiers, robot_case_split_States, verbose);
		    }
		    else
		    {
			encoding_context.m_clause_generator->generate_Disequality(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(layer)), vertex_id);
		    }
		}
	    }
	}

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_initial_arrangement.get_RobotLocation(robot_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(0)), vertex_id, verbose);
	}
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int vertex_id = m_goal_arrangement.get_RobotLocation(robot_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&robot_location, sIntegerIndex(robot_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), vertex_id, verbose);
	}
    }


    void sMultirobotInstance::to_Stream_HeuristicBijectionCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent, bool verbose) const
    {
	encoding_context.switchTo_AdvancedGeneratingMode();
	to_Stream_HeuristicDifferentialCNFsat(fw, encoding_context, indent, verbose);
    }


    void sMultirobotInstance::to_Stream_HeuristicAdvancedCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &sUNUSED(indent), bool verbose) const
    {
	encoding_context.switchTo_AdvancedGeneratingMode();

	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	m_environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	int Clause_cnt = 0;
	int aux_Variable_cnt = 0;
 
	int N_Vertices = m_environment.get_VertexCount();
	int N_Robots = m_initial_arrangement.get_RobotCount();

	sIndexableStateIdentifier vertex_occupancy(&encoding_context.m_variable_store, "vertex_occupancy", N_Robots + 1, sIntegerScope(0, N_Vertices), sIntegerScope(0, encoding_context.m_N_Layers - 1));
	encoding_context.m_vertex_occupancy = vertex_occupancy;
	encoding_context.register_TranslateIdentifier(encoding_context.m_vertex_occupancy);

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
	    }
	}

	encoding_context.m_transition_Actions.resize(N_Vertices);
	
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    sIndexableStateIdentifier transition_action(&encoding_context.m_variable_store, "transition_action-" + sInt_32_to_String(vertex_id), 2 * m_environment.get_Vertex(vertex_id)->calc_NeighborCount() + 1, sIntegerScope(0, encoding_context.m_N_Layers - 1)); 
	    encoding_context.m_transition_Actions[vertex_id] = transition_action;
	    encoding_context.register_TranslateIdentifier(encoding_context.m_transition_Actions[vertex_id]);

	    for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	    {		    
		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&transition_action, sIntegerIndex(layer)));

		for (int robot_id = 1; robot_id < N_Robots; ++robot_id)
		{
		    if (all_pairs_Distances[m_initial_arrangement.get_RobotLocation(robot_id)][vertex_id] > layer || all_pairs_Distances[vertex_id][m_goal_arrangement.get_RobotLocation(robot_id)] > encoding_context.m_N_Layers - layer)
		    {
			Clause_cnt += encoding_context.m_clause_generator->count_Disequality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)), robot_id);
		    }
		}
	    }
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{	
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {
		int N_vertex_Neighbors = m_environment.get_Vertex(vertex_id)->calc_NeighborCount();
		const sVertex::Neighbors_list &out_Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;

		int out_neighbor_order = 0;

		for (sVertex::Neighbors_list::const_iterator out_neighbor = out_Neighbors.begin(); out_neighbor != out_Neighbors.end(); ++out_neighbor)
		{
		    int neighbor_id = (*out_neighbor)->m_target->m_id;
		    int N_neighbor_Neighbors = m_environment.get_Vertex(neighbor_id)->calc_NeighborCount();
		    int in_neighbor_order = m_environment.get_Vertex(neighbor_id)->calc_NeighborOrder(vertex_id);

		    Clause_cnt += encoding_context.m_clause_generator->count_ConditionalEquality(aux_Variable_cnt,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
												 out_neighbor_order,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
												 N_neighbor_Neighbors + in_neighbor_order,
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer)),
												 0,
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
												 0,
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer + 1)),
												 sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
		    
		    Clause_cnt += encoding_context.m_clause_generator->count_ConditionalEquality(aux_Variable_cnt,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
												 N_vertex_Neighbors + out_neighbor_order,
												 sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
												 in_neighbor_order);

		    ++out_neighbor_order;
		}
		
		Clause_cnt += encoding_context.m_clause_generator->count_ConditionalEquality(aux_Variable_cnt,
											     sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
											     2 * N_vertex_Neighbors,
											     sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
											     sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
	    }
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(0)), robot_id);
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), robot_id);
	}

	int N_cnf_Variables = encoding_context.m_variable_store.get_Last_CNFVariable() + aux_Variable_cnt - 1;

	fprintf(fw, "c %s : multirobot inverse SAT encoding\n", sPRODUCT);
	fprintf(fw, "c %s\n", sCOPYRIGHT);
	fprintf(fw, "c number of layers = %d\n", encoding_context.m_N_Layers);
	fprintf(fw, "c number of clauses / number of variables = %.3f\n", (double)Clause_cnt / N_cnf_Variables);
	fprintf(fw, "c number of visible variables = %d\n", N_cnf_Variables - aux_Variable_cnt);
	fprintf(fw, "c number of hidden variables = %d\n", aux_Variable_cnt);
	to_Stream(fw, "c ");
	fprintf(fw, "p cnf %d %d\n", N_cnf_Variables, Clause_cnt);

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	    {
		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)), verbose);
	    }
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	    {		    
		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)), verbose);

		for (int robot_id = 1; robot_id < N_Robots; ++robot_id)
		{
		    if (all_pairs_Distances[m_initial_arrangement.get_RobotLocation(robot_id)][vertex_id] > layer || all_pairs_Distances[vertex_id][m_goal_arrangement.get_RobotLocation(robot_id)] > encoding_context.m_N_Layers - layer)
		    {
			encoding_context.m_clause_generator->generate_Disequality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)), robot_id);
		    }
		}
	    }
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{	
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {
		int N_vertex_Neighbors = m_environment.get_Vertex(vertex_id)->calc_NeighborCount();
		const sVertex::Neighbors_list &out_Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;

		int out_neighbor_order = 0;

		for (sVertex::Neighbors_list::const_iterator out_neighbor = out_Neighbors.begin(); out_neighbor != out_Neighbors.end(); ++out_neighbor)
		{
		    int neighbor_id = (*out_neighbor)->m_target->m_id;
		    int N_neighbor_Neighbors = m_environment.get_Vertex(neighbor_id)->calc_NeighborCount();
		    int in_neighbor_order = m_environment.get_Vertex(neighbor_id)->calc_NeighborOrder(vertex_id);

		    encoding_context.m_clause_generator->generate_ConditionalEquality(fw,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
										      out_neighbor_order,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
										      N_neighbor_Neighbors + in_neighbor_order,
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer)),
										      0,
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
										      0,
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer + 1)),
										      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)),
										      verbose);
		    
		    encoding_context.m_clause_generator->generate_ConditionalEquality(fw,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
										      N_vertex_Neighbors + out_neighbor_order,
										      sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[neighbor_id], sIntegerIndex(layer)),
										      in_neighbor_order,
										      verbose);
		    
		    ++out_neighbor_order;
		}
		encoding_context.m_clause_generator->generate_ConditionalEquality(fw,
										  sSpecifiedStateIdentifier(&encoding_context.m_transition_Actions[vertex_id], sIntegerIndex(layer)),
										  2 * N_vertex_Neighbors,
										  sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
										  sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)),
										  verbose);
	    }
	}
	
	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(0)), robot_id, verbose);
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), robot_id, verbose);
	} 
    }


    void sMultirobotInstance::to_Stream_PuzzleCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &sUNUSED(indent), bool verbose) const
    {
	encoding_context.switchTo_AdvancedGeneratingMode();

	int Clause_cnt = 0;
	int aux_Variable_cnt = 0;

	int N_Vertices = m_environment.get_VertexCount();
	int N_Robots = m_initial_arrangement.get_RobotCount();
	int N_Edges = m_environment.get_EdgeCount();

	sIndexableStateIdentifier vertex_occupancy(&encoding_context.m_variable_store, "vertex_occupancy", N_Robots + 1, sIntegerScope(0, N_Vertices), sIntegerScope(0, encoding_context.m_N_Layers - 1));
	encoding_context.m_vertex_occupancy = vertex_occupancy;
	encoding_context.register_TranslateIdentifier(encoding_context.m_vertex_occupancy);

	sIndexableStateIdentifier transition_action(&encoding_context.m_variable_store, "transition_action", 1 + 2 * N_Edges, sIntegerScope(0, encoding_context.m_N_Layers - 1));
	encoding_context.m_transition_action = transition_action;
	encoding_context.register_TranslateIdentifier(encoding_context.m_transition_action);

	for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	{
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {

		Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));
	    }
	    Clause_cnt += encoding_context.m_clause_generator->count_Alignment(aux_Variable_cnt, sSpecifiedStateIdentifier(&transition_action, sIntegerIndex(layer)));
	}
         
	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{	
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {
		const sVertex::Neighbors_list &out_Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;

		for (sVertex::Neighbors_list::const_iterator out_neighbor = out_Neighbors.begin(); out_neighbor != out_Neighbors.end(); ++out_neighbor)
		{
		    int neighbor_id = (*out_neighbor)->m_target->m_id;
		    int action_id;

		    if ((*out_neighbor)->m_edge->m_arc_uv.m_target->m_id == vertex_id)
		    {
			action_id = (*out_neighbor)->m_edge->m_id;
		    }
		    else
		    {
			action_id = (*out_neighbor)->m_edge->m_id + N_Edges;
		    }

		    sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_A;
		    sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_B;

		    occupancy_Identifiers_A.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer + 1)));
		    occupancy_Identifiers_B.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));

		    for (int remain_vertex_id = 0; remain_vertex_id < N_Vertices; ++remain_vertex_id)
		    {
			if (remain_vertex_id != vertex_id && remain_vertex_id != neighbor_id)
			{
			    occupancy_Identifiers_A.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer + 1)));
			    occupancy_Identifiers_B.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer)));
			}
		    }
		    
		    Clause_cnt += encoding_context.m_clause_generator->count_LargeConditionalEquality(aux_Variable_cnt,
												      sSpecifiedStateIdentifier(&encoding_context.m_transition_action, sIntegerIndex(layer)),
												      action_id,
												      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
												      0,
												      sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer)),
												      0,
												      occupancy_Identifiers_A,
												      occupancy_Identifiers_B);
		}
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_A;
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_B;

		for (int remain_vertex_id = 0; remain_vertex_id < N_Vertices; ++remain_vertex_id)
		{
		    occupancy_Identifiers_A.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer + 1)));
		    occupancy_Identifiers_B.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer)));
		}

		Clause_cnt += encoding_context.m_clause_generator->count_LargeConditionalEquality(aux_Variable_cnt,
												  sSpecifiedStateIdentifier(&encoding_context.m_transition_action, sIntegerIndex(layer)),
												  2 * N_Edges,
												  occupancy_Identifiers_A,
												  occupancy_Identifiers_B);
	    }
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(0)), robot_id);
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);
	    Clause_cnt += encoding_context.m_clause_generator->count_Equality(aux_Variable_cnt, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), robot_id);
	}

	int N_cnf_Variables = encoding_context.m_variable_store.get_Last_CNFVariable() + aux_Variable_cnt - 1;

	fprintf(fw, "c %s : multirobot inverse SAT encoding\n", sPRODUCT);
	fprintf(fw, "c %s\n", sCOPYRIGHT);
	fprintf(fw, "c number of layers = %d\n", encoding_context.m_N_Layers);
	fprintf(fw, "c number of clauses / number of variables = %.3f\n", (double)Clause_cnt / N_cnf_Variables);
	fprintf(fw, "c number of visible variables = %d\n", N_cnf_Variables - aux_Variable_cnt);
	fprintf(fw, "c number of hidden variables = %d\n", aux_Variable_cnt);
	to_Stream(fw, "c ");
	fprintf(fw, "p cnf %d %d\n", N_cnf_Variables, Clause_cnt);

	for (int layer = 0; layer < encoding_context.m_N_Layers; ++layer)
	{
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {

		encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)), verbose);
	    }
	    encoding_context.m_clause_generator->generate_Alignment(fw, sSpecifiedStateIdentifier(&transition_action, sIntegerIndex(layer)), verbose);
	}

	for (int layer = 0; layer < encoding_context.m_N_Layers - 1; ++layer)
	{	
	    for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	    {
		const sVertex::Neighbors_list &out_Neighbors =  m_environment.get_Vertex(vertex_id)->m_Neighbors;

		for (sVertex::Neighbors_list::const_iterator out_neighbor = out_Neighbors.begin(); out_neighbor != out_Neighbors.end(); ++out_neighbor)
		{
		    int neighbor_id = (*out_neighbor)->m_target->m_id;

		    int action_id;
		    if ((*out_neighbor)->m_edge->m_arc_uv.m_target->m_id == vertex_id)
		    {
			action_id = (*out_neighbor)->m_edge->m_id;
		    }
		    else
		    {
			action_id = (*out_neighbor)->m_edge->m_id + N_Edges;
		    }

		    sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_A;
		    sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_B;

		    occupancy_Identifiers_A.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer + 1)));
		    occupancy_Identifiers_B.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer)));

		    for (int remain_vertex_id = 0; remain_vertex_id < N_Vertices; ++remain_vertex_id)
		    {
			if (remain_vertex_id != vertex_id && remain_vertex_id != neighbor_id)
			{
			    occupancy_Identifiers_A.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer + 1)));
			    occupancy_Identifiers_B.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer)));
			}
		    }
		    
		    encoding_context.m_clause_generator->generate_LargeConditionalEquality(fw,
											   sSpecifiedStateIdentifier(&encoding_context.m_transition_action, sIntegerIndex(layer)),
											   action_id,
											   sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(layer + 1)),
											   0,
											   sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(neighbor_id), sIntegerIndex(layer)),
											   0,
											   occupancy_Identifiers_A,
											   occupancy_Identifiers_B,
											   verbose);
		}
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_A;
		sStateClauseGenerator::SpecifiedStateIdentifiers_vector occupancy_Identifiers_B;

		for (int remain_vertex_id = 0; remain_vertex_id < N_Vertices; ++remain_vertex_id)
		{
		    occupancy_Identifiers_A.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer + 1)));
		    occupancy_Identifiers_B.push_back(sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(remain_vertex_id), sIntegerIndex(layer)));
		}

		encoding_context.m_clause_generator->generate_LargeConditionalEquality(fw,
										       sSpecifiedStateIdentifier(&encoding_context.m_transition_action, sIntegerIndex(layer)),
										       2 * N_Edges,
										       occupancy_Identifiers_A,
										       occupancy_Identifiers_B,
										       verbose);
	    }
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_initial_arrangement.get_VertexOccupancy(vertex_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(0)), robot_id, verbose);
	}

	for (int vertex_id = 0; vertex_id < N_Vertices; ++vertex_id)
	{
	    int robot_id = m_goal_arrangement.get_VertexOccupancy(vertex_id);
	    encoding_context.m_clause_generator->generate_Equality(fw, sSpecifiedStateIdentifier(&vertex_occupancy, sIntegerIndex(vertex_id), sIntegerIndex(encoding_context.m_N_Layers - 1)), robot_id, verbose);
	} 	
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolution

    const sMultirobotSolution::Move sMultirobotSolution::UNDEFINED_MOVE = sMultirobotSolution::Move(0, 0, 0);


/*----------------------------------------------------------------------------*/

    sMultirobotSolution::Move::Move(int robot_id, int src_vrtx_id, int dest_vrtx_id)
	: m_robot_id(robot_id),
	  m_src_vrtx_id(src_vrtx_id),
	  m_dest_vrtx_id(dest_vrtx_id)
    {
	// nothing
    }


    bool sMultirobotSolution::Move::is_Undefined(void) const
    {
	return (m_robot_id <= 0);
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolution::Step::Step(int time)
	: m_time(time)
    {
	// nothing
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolution::sMultirobotSolution()
	: m_Moves_cnt(0)
    {
	// nothing
    }


    sMultirobotSolution::sMultirobotSolution(int start_step, const sMultirobotSolution &sub_solution)
	: m_Moves_cnt(0)
    {
	int N_Steps = sub_solution.m_Steps.size();
	for (int i = 0; i < N_Steps; ++i)
	{
	    const Step &step = sub_solution.m_Steps[i];

	    for (Moves_list::const_iterator move = step.m_Moves.begin(); move != step.m_Moves.end(); ++move)
	    {
		add_Move(i + start_step, *move);
	    }
	}	
    }


    sMultirobotSolution::sMultirobotSolution(const sMultirobotSolution &sub_solution_1, const sMultirobotSolution sub_solution_2)
	: m_Moves_cnt(0)
    {
	int N_Steps_1 = sub_solution_1.m_Steps.size();
	for (int i = 0; i < N_Steps_1; ++i)
	{
	    const Step &step = sub_solution_1.m_Steps[i];

	    for (Moves_list::const_iterator move = step.m_Moves.begin(); move != step.m_Moves.end(); ++move)
	    {
		add_Move(i, *move);
	    }
	}

	int N_Steps_2 = sub_solution_2.m_Steps.size();
	for (int i = 0; i < N_Steps_2; ++i)
	{
	    const Step &step = sub_solution_2.m_Steps[i];

	    for (Moves_list::const_iterator move = step.m_Moves.begin(); move != step.m_Moves.end(); ++move)
	    {
		add_Move(i, *move);
	    }
	}
    }


    sMultirobotSolution::sMultirobotSolution(const sMultirobotSolution &multirobot_solution)
	: m_Moves_cnt(multirobot_solution.m_Moves_cnt)
	, m_Steps(multirobot_solution.m_Steps)
    {
	// nothing
    }


    const sMultirobotSolution& sMultirobotSolution::operator=(const sMultirobotSolution &multirobot_solution)
    {
	m_Moves_cnt = multirobot_solution.m_Moves_cnt;
	m_Steps = multirobot_solution.m_Steps;

	return *this;
    }


    int sMultirobotSolution::get_MoveCount(void) const
    {
	return m_Moves_cnt;
    }


    int sMultirobotSolution::get_StepCount(void) const
    {
	return m_Steps.size();
    }


    void sMultirobotSolution::add_Move(int time, const Move &move)
    {
	int push_cnt = time - m_Steps.size();

	while (push_cnt-- >= 0)
	{
	    m_Steps.push_back(Step(m_Steps.size()));
	}
	m_Steps[time].m_Moves.push_back(move);
	++m_Moves_cnt;
    }


    int sMultirobotSolution::calc_EmptySteps(void) const
    {
	int N_empty_Steps = 0;
	int N_Steps = m_Steps.size();
	for (int i = 0; i < N_Steps; ++i)
	{
	    const Step &step = m_Steps[i];
	    if (step.m_Moves.empty())
	    {
		++N_empty_Steps;
	    }
	}
	return N_empty_Steps;
    }


    void sMultirobotSolution::remove_EmptySteps(void)
    {
	int time = 0;
	Steps_vector clean_Steps;

	int N_Steps = m_Steps.size();
	for (int i = 0; i < N_Steps; ++i)
	{
	    Step &step = m_Steps[i];

	    if (!step.m_Moves.empty())
	    {
		step.m_time = time++;
		clean_Steps.push_back(step);
	    }
	}
	m_Steps = clean_Steps;
    }


    sMultirobotSolution sMultirobotSolution::extract_Subsolution(int start_step, int final_step) const
    {
	sMultirobotSolution subsolution;

	for (int i = start_step; i <= final_step; ++i)
	{
	    const Step &step = m_Steps[i];

	    for (Moves_list::const_iterator move = step.m_Moves.begin(); move != step.m_Moves.end(); ++move)
	    {
		subsolution.add_Move(i, *move);
	    }
	}	

	return subsolution;
    }


    void sMultirobotSolution::execute_Solution(const sRobotArrangement &initial_arrangement,
					       sRobotArrangement       &final_arrangement,
					       int                      N_Steps) const
    {
	final_arrangement = initial_arrangement;

	if (N_Steps == N_STEPS_UNDEFINED)
	{
	    for (sMultirobotSolution::Steps_vector::const_iterator step =  m_Steps.begin(); step != m_Steps.end(); ++step)
	    {
		for (sMultirobotSolution::Moves_list::const_iterator move = step->m_Moves.begin(); move != step->m_Moves.end(); ++move)
		{
		    final_arrangement.move_Robot(move->m_robot_id, move->m_dest_vrtx_id);
		}	    
	    }
	}
	else
	{
	    for (sMultirobotSolution::Steps_vector::const_iterator step =  m_Steps.begin(); step != m_Steps.end(); ++step)
	    {
		for (sMultirobotSolution::Moves_list::const_iterator move = step->m_Moves.begin(); move != step->m_Moves.end(); ++move)
		{
		    final_arrangement.move_Robot(move->m_robot_id, move->m_dest_vrtx_id);
		}	    
		if (--N_Steps <= 0)
		{
		    return;
		}
	    }
	}
    }


    void sMultirobotSolution::to_Screen(const sString &indent) const
    {
	to_Stream(stdout, indent);
    }


    void sMultirobotSolution::to_Stream(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%sMulirobot solution: (|moves| = %d, paralellism = %.3f) [\n", indent.c_str(), m_Moves_cnt, (double)m_Moves_cnt / m_Steps.size());

	int N_Steps = m_Steps.size();
	for (int i = 0; i < N_Steps; ++i)
	{
	    const Step &step = m_Steps[i];
	    fprintf(fw, "%s%sStep %d: ", indent.c_str(), sRELOC_INDENT.c_str(), step.m_time);

	    for (Moves_list::const_iterator move = step.m_Moves.begin(); move != step.m_Moves.end(); ++move)
	    {
		fprintf(fw, "%d#%d->%d ", move->m_robot_id, move->m_src_vrtx_id, move->m_dest_vrtx_id);
	    }
	    fprintf(fw, "\n");
	}

	fprintf(fw, "%s]\n", indent.c_str());
    }


    sResult sMultirobotSolution::to_File_multirobot(const sString &filename, const sString &indent) const
    {
	FILE *fw;

	if ((fw = fopen(filename.c_str(), "w")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_OPEN_ERROR;
	}
	
	to_Stream_multirobot(fw, indent);
	fclose(fw);

	return sRESULT_SUCCESS;
    }


    void sMultirobotSolution::to_Stream_multirobot(FILE *fw, const sString &indent) const
    {
	fprintf(fw, "%sFine solution\n", indent.c_str());
	fprintf(fw, "%sLength:%d\n", indent.c_str(), get_MoveCount());
	int N_Steps = m_Steps.size();
	for (int i = 0; i < N_Steps; ++i)
	{
	    const Step &step = m_Steps[i];
	    for (Moves_list::const_iterator move = step.m_Moves.begin(); move != step.m_Moves.end(); ++move)
	    {
		fprintf(fw, "%s%d # %d ---> %d (%d)\n", indent.c_str(), move->m_robot_id, move->m_src_vrtx_id, move->m_dest_vrtx_id, i);
	    }
	}
    }


    sResult sMultirobotSolution::from_File_multirobot(const sString &filename)
    {
	sResult result;
	FILE *fr;

	if ((fr = fopen(filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_OPEN_ERROR;
	}
	
	result = from_Stream_multirobot(fr);
	if (sFAILED(result))
	{
	    fclose(fr);
	    return result;
	}
	fclose(fr);

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolution::from_Stream_multirobot(FILE *fr)
    {
	int N_Moves;
	int c = fgetc(fr);

	while (c != 'F')
	{
	    if (c != '\n')
	    {
		while(fgetc(fr) != '\n');
	    }
	    c = fgetc(fr);
	}

	fscanf(fr, "ine solution\nLength:%d\n", &N_Moves);
        #ifdef sDEBUG
	printf("Length:%d\n", N_Moves);
	#endif

	for (int i = 0; i < N_Moves; ++i)
	{
	    int robot_id, src_vertex_id, dest_vertex_id, step;
	    fscanf(fr, "%d # %d ---> %d (%d)\n", &robot_id, &src_vertex_id, &dest_vertex_id, &step);
	    #ifdef sDEBUG
	    printf("%d # %d ---> %d (%d)\n", robot_id, src_vertex_id, dest_vertex_id, step);
	    #endif
	    add_Move(step, Move(robot_id, src_vertex_id, dest_vertex_id));
	}

	return sRESULT_SUCCESS;
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolver

    sMultirobotSolver::sMultirobotSolver()
    {
	// nothing
    }


    void sMultirobotSolver::setup_Solver(void)
    {
	// nothing;
    }


    void sMultirobotSolver::setup_Instance(const sMultirobotInstance &multirobot_instnace)
    {
	m_multirobot_instance = multirobot_instnace;
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolver_ID

    sMultirobotSolver_ID::sMultirobotSolver_ID()
	: m_max_depth_limit(0),
	  m_finished_Robot_cnt(0)
    {
	// nothing
    }


    void sMultirobotSolver_ID::setup_Solver(int max_depth_limit)
    {
	sMultirobotSolver::setup_Solver();
	m_max_depth_limit = max_depth_limit;
    }


    void sMultirobotSolver_ID::setup_Instance(const sMultirobotInstance &multirobot_instance)
    {
	sMultirobotSolver::setup_Instance(multirobot_instance);
	m_current_arrangement = multirobot_instance.m_initial_arrangement;

	m_finished_Robot_cnt = 0;

	int N_Robots = m_current_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    if (m_current_arrangement.get_RobotLocation(robot_id) == m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id))
	    {
		++m_finished_Robot_cnt;
	    }
	}
    }

    
    bool sMultirobotSolver_ID::solve_Instance(sMultirobotSolution &multirobot_solution)
    {
	for (int depth_limit = 1; m_max_depth_limit < 0 || depth_limit <= m_max_depth_limit; ++depth_limit)
	{
	    bool solution_found = search_Goal(0, depth_limit, multirobot_solution);

	    if (solution_found)
	    {
		return true;
	    }
	}

	return false;
    }


    bool sMultirobotSolver_ID::search_Goal(int current_depth, int depth_limit, sMultirobotSolution &multirobot_solution)
    {
#ifdef sSTATISTICS
	static sUInt_64 search_Step_cnt = 0;
	++search_Step_cnt;
#endif    

#ifdef sVERBOSE	
	static sUInt_64 verbose_period = 65536;
	static sUInt_64 Verbose_cnt = 0;

	if (Verbose_cnt++ >= verbose_period)
	{
	    printf("Current depth/limit (steps): ");
	    for (int i = 0; i < depth_limit; ++i)
	    {
		printf(".");
	    }
	    printf(" %d/%d (%llu)\n", current_depth, depth_limit, search_Step_cnt);

	    Verbose_cnt = 0;
	    verbose_period *= 1.8;
	    ++verbose_period;
	}
#endif

	if (m_finished_Robot_cnt == m_current_arrangement.get_RobotCount())
	{
	    return true;
	}
	else
	{
	    if (current_depth < depth_limit)
	    {
		int N_Robots = m_current_arrangement.get_RobotCount();
		for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
		{
		    int robot_vertex_id = m_current_arrangement.get_RobotLocation(robot_id);
		    sVertex* robot_vertex = m_multirobot_instance.m_environment.get_Vertex(robot_vertex_id);
		    
		    for (sVertex::Neighbors_list::const_iterator neighbor = robot_vertex->m_Neighbors.begin(); neighbor != robot_vertex->m_Neighbors.end(); ++neighbor)
		    {
			int neighbor_vertex_id = (*neighbor)->m_target->m_id;
			
			if (m_current_arrangement.get_VertexOccupancy(neighbor_vertex_id) == sRobotArrangement::VACANT_VERTEX)
			{
			    conduct_Move(robot_id, robot_vertex_id, neighbor_vertex_id);			    
			    bool result = search_Goal(current_depth + 1, depth_limit, multirobot_solution);
			    
			    if (result)
			    {
				multirobot_solution.add_Move(current_depth, sMultirobotSolution::Move(robot_id, robot_vertex_id, neighbor_vertex_id));
				return true;
			    }
			    else
			    {
				conduct_Move(robot_id, neighbor_vertex_id, robot_vertex_id);
			    }		 
			}
		    }
		}
		return false;
	    }
	    else
	    {
		return false;
	    }
	}

	sASSERT(false);
	return false;
    }


    void sMultirobotSolver_ID::conduct_Move(int robot_id, int src_vrtx_id, int dest_vrtx_id)
    {
	const sRobotArrangement &goal_arrangement = m_multirobot_instance.m_goal_arrangement;

	if (src_vrtx_id == goal_arrangement.get_RobotLocation(robot_id) && src_vrtx_id != dest_vrtx_id)
	{
	    --m_finished_Robot_cnt;
	}
	else
	{
	    if (dest_vrtx_id == goal_arrangement.get_RobotLocation(robot_id) && dest_vrtx_id != src_vrtx_id)
	    {
		++m_finished_Robot_cnt;
	    }
	}
	m_current_arrangement.move_Robot(robot_id, dest_vrtx_id);
    }


/*----------------------------------------------------------------------------*/
// sAstarHeuristic_Trivial

    sAstarHeuristic_Trivial::sAstarHeuristic_Trivial()
    {
	// nothing
    }


    void sAstarHeuristic_Trivial::setup_Instance(const sMultirobotInstance &sUNUSED(multirobot_instance))
    {
	// nothing
    }


    int sAstarHeuristic_Trivial::calc_Heuristic(int sUNUSED(source_vertex_id), int sUNUSED(destination_vertex_id))
    {
	return 0;
    }


    int sAstarHeuristic_Trivial::calc_Heuristic(const sRobotArrangement &sUNUSED(current_arrangement), const sRobotArrangement &sUNUSED(goal_arrangement))
    {
	return 0;
    }


/*----------------------------------------------------------------------------*/
// sAstarHeuristic_Placement

    sAstarHeuristic_Placement::sAstarHeuristic_Placement()
    {
	// nothing
    }


    void sAstarHeuristic_Placement::setup_Instance(const sMultirobotInstance &sUNUSED(multirobot_instance))
    {
	// nothing
    }


    int sAstarHeuristic_Placement::calc_Heuristic(int source_vertex_id, int destination_vertex_id)
    {
	if (source_vertex_id == destination_vertex_id)
	{
	    return 0;
	}
	else
	{
	    return 1;
	}
    }


    int sAstarHeuristic_Placement::calc_Heuristic(const sRobotArrangement &current_arrangement, const sRobotArrangement &goal_arrangement)
    {
	int placed_Robot_cnt = 0;

	int N_Robots = current_arrangement.get_RobotCount();
	for (int r_id = 1; r_id <= N_Robots; ++r_id)
	{
	    if (current_arrangement.get_RobotLocation(r_id) == goal_arrangement.get_RobotLocation(r_id))
	    {
		++placed_Robot_cnt;
	    }
	}

	return N_Robots - placed_Robot_cnt;
    }


/*----------------------------------------------------------------------------*/
// sAstarHeuristic_Distance

    sAstarHeuristic_Distance::sAstarHeuristic_Distance()
    {
	// nothing
    }


    void sAstarHeuristic_Distance::setup_Instance(const sMultirobotInstance &multirobot_instance)
    {
	const sUndirectedGraph &environment = multirobot_instance.m_environment;
	int N_Vertices = environment.get_VertexCount();

	for (int v_id = 0; v_id < N_Vertices; ++v_id)
	{
	    MatrixRow_vector matrix_Row;
	    environment.calc_SingleSourceShortestPaths(v_id, matrix_Row);
	    m_distance_Matrix.push_back(matrix_Row);
	}
    }


    int sAstarHeuristic_Distance::calc_Heuristic(int source_vertex_id, int destination_vertex_id)
    {
	return m_distance_Matrix[source_vertex_id][destination_vertex_id];
    }


    int sAstarHeuristic_Distance::calc_Heuristic(const sRobotArrangement &current_arrangement, const sRobotArrangement &goal_arrangement)
    {
	int total_distance = 0;
	int N_Robots = current_arrangement.get_RobotCount();

	for (int r_id = 1; r_id <= N_Robots; ++r_id)
	{
	    total_distance += m_distance_Matrix[current_arrangement.get_RobotLocation(r_id)][goal_arrangement.get_RobotLocation(r_id)];
	}

	return total_distance;
    }


    void sAstarHeuristic_Distance::to_String_matrix(const sString &indent) const
    {
	sASSERT(!m_distance_Matrix.empty());

	printf("%sDistance matrix:\n", indent.c_str());
	printf("%s\t", indent.c_str());

	for (int column = 0; column < m_distance_Matrix[0].size(); ++column)
	{
	    printf("%d\t", column);
	}
	printf("\n");

	for (int row = 0; row < m_distance_Matrix.size(); ++row)
	{
	    printf("%s%d:\t", indent.c_str(), row);
	    for (MatrixRow_vector::const_iterator cell = m_distance_Matrix[row].begin(); cell != m_distance_Matrix[row].end(); ++cell)
	    {
		printf("%d\t", *cell);
	    }
	    printf("\n");
	}
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolver_Astar

    static const sMultirobotSolver_Astar::ProductionMove UNDEFINED_PRODUCTION_MODE = sMultirobotSolver_Astar::ProductionMove(NULL, sMultirobotSolution::UNDEFINED_MOVE);


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_Astar::ProductionMove::ProductionMove(ProductionMove *prev, sMultirobotSolution::Move move)
	: m_prev(prev),
	  m_move(move)
    {
	// nothing
    }

/*----------------------------------------------------------------------------*/


    sMultirobotSolver_Astar::StateRecord::StateRecord(const sRobotArrangement &robot_arrangement)
	: m_cost(0),
	  m_production_move(NULL),
	  m_robot_arrangement(robot_arrangement)
    {
	// nothing
    }


    sMultirobotSolver_Astar::StateRecord::StateRecord(int cost, ProductionMove *production_move, const sRobotArrangement &robot_arrangement)
	: m_cost(cost),
	  m_production_move(production_move),
	  m_robot_arrangement(robot_arrangement)
    {
	// nothing
    }


    bool sMultirobotSolver_Astar::StateRecord::operator==(const StateRecord &state_record) const
    {
	return (m_robot_arrangement == state_record.m_robot_arrangement);
    }


    bool sMultirobotSolver_Astar::StateRecord::operator<(const StateRecord &state_record) const
    {
	return (m_robot_arrangement < state_record.m_robot_arrangement);
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_Astar::sMultirobotSolver_Astar()
	: m_heuristic(NULL)
    {
	// nothing
    }


    void sMultirobotSolver_Astar::setup_Solver(sAstarHeuristic *heuristic)
    {
	sMultirobotSolver::setup_Solver();
	m_heuristic = heuristic;
    } 


    void sMultirobotSolver_Astar::setup_Instance(const sMultirobotInstance &multirobot_instance)
    {
	sMultirobotSolver::setup_Instance(multirobot_instance);
	m_heuristic->setup_Instance(multirobot_instance);
    }


    bool sMultirobotSolver_Astar::solve_Instance(sMultirobotSolution &multirobot_solution)
    {
	int first_state_cost = m_heuristic->calc_Heuristic(m_multirobot_instance.m_initial_arrangement, m_multirobot_instance.m_goal_arrangement);
	StateRecords_set::iterator first_state_iter = m_open_state_Recs.insert(StateRecord(first_state_cost, NULL, m_multirobot_instance.m_initial_arrangement)).first;
	m_expansion_Queue.insert(StateRecords_multimap::value_type(first_state_cost, first_state_iter));

	while (!m_open_state_Recs.empty())
	{
#ifdef sSTATISTICS
	    static sUInt_64 search_Step_cnt = 0;
	    ++search_Step_cnt;
#endif    

#ifdef sVERBOSE	
	    static sUInt_64 verbose_period = 65536;
	    static sUInt_64 Verbose_cnt = 0;

	    if (Verbose_cnt++ >= verbose_period)
	    {
		printf("Open/close/exp size (steps): %lu/%lu/%lu (%llu)\n", m_open_state_Recs.size(), m_close_state_Recs.size(), m_expansion_Queue.size(), search_Step_cnt);
		
		Verbose_cnt = 0;
		verbose_period *= 1.8;
		++verbose_period;
	    }
#endif
	    StateRecords_multimap::iterator expand_record = m_expansion_Queue.begin();

	    if (expand_Arrangement(*expand_record->second, multirobot_solution))
	    {
		return true;
	    }
	    m_open_state_Recs.erase(expand_record->second);
	    m_expansion_Queue.erase(expand_record);
	}

	return false;
    }


    bool sMultirobotSolver_Astar::expand_Arrangement(const StateRecord &current_state_rec, sMultirobotSolution &multirobot_solution)
    {
	int N_Robots = current_state_rec.m_robot_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int robot_vertex_id = current_state_rec.m_robot_arrangement.get_RobotLocation(robot_id);
	    sVertex* robot_vertex = m_multirobot_instance.m_environment.get_Vertex(robot_vertex_id);
	    
	    for (sVertex::Neighbors_list::const_iterator neighbor = robot_vertex->m_Neighbors.begin(); neighbor != robot_vertex->m_Neighbors.end(); ++neighbor)
	    {
		int neighbor_vertex_id = (*neighbor)->m_target->m_id;
		
		if (current_state_rec.m_robot_arrangement.get_VertexOccupancy(neighbor_vertex_id) == sRobotArrangement::VACANT_VERTEX)
		{
		    sRobotArrangement next_arrangement;
		    conduct_Move(robot_id, robot_vertex_id, neighbor_vertex_id, current_state_rec.m_robot_arrangement, next_arrangement);
		    if (next_arrangement == m_multirobot_instance.m_goal_arrangement)
		    {
			int time = collect_Solution(current_state_rec, multirobot_solution);
			multirobot_solution.add_Move(time, sMultirobotSolution::Move(robot_id, robot_vertex_id, neighbor_vertex_id));
			return true;
		    }
		    else
		    {
			if (m_open_state_Recs.find(StateRecord(next_arrangement)) == m_open_state_Recs.end() && m_close_state_Recs.find(StateRecord(next_arrangement)) == m_close_state_Recs.end())
			{
			    m_production_Moves.push_back(ProductionMove(current_state_rec.m_production_move, sMultirobotSolution::Move(robot_id, robot_vertex_id, neighbor_vertex_id)));
			    int next_state_cost = m_heuristic->calc_Heuristic(next_arrangement, m_multirobot_instance.m_goal_arrangement) + current_state_rec.m_cost;
			    StateRecords_set::iterator next_state_iter = m_open_state_Recs.insert(StateRecord(next_state_cost, &m_production_Moves.back(), next_arrangement)).first;

			    m_expansion_Queue.insert(StateRecords_multimap::value_type(next_state_cost, next_state_iter));
			}
		    }
		}
	    }
	}
	m_close_state_Recs.insert(current_state_rec);

	/*
	printf("Open list:\n");
	for (StateRecords_set::const_iterator open = m_open_state_Recs.begin(); open != m_open_state_Recs.end(); ++open)
	{
	    open->m_robot_arrangement.to_Screen();
	}
	printf("Close list:\n");
	for (StateRecords_set::const_iterator close = m_close_state_Recs.begin(); close != m_close_state_Recs.end(); ++close)
	{
	    close->m_robot_arrangement.to_Screen();
	}
	printf("Expansion queue:\n");
	for (StateRecords_multimap::const_iterator elem = m_expansion_queue.begin(); elem != m_expansion_queue.end(); ++elem)
	{
	    elem->second->m_robot_arrangement.to_Screen();
	}
	*/
	return false;
    }


    int sMultirobotSolver_Astar::collect_Solution(const StateRecord &current_state_rec, sMultirobotSolution &multirobot_solution)
    {
	ProductionMove *production_move = current_state_rec.m_production_move;
	vector<ProductionMove*> production_Moves;

	for (; production_move != NULL; production_move = production_move->m_prev)
	{
	    production_Moves.push_back(production_move);
	}

	int time = 0;
	for (vector<ProductionMove*>::const_reverse_iterator prod_move = production_Moves.rbegin(); prod_move != production_Moves.rend(); ++prod_move, ++time)
	{
	    multirobot_solution.add_Move(time, (*prod_move)->m_move);
	}
	return time;
    }


    void sMultirobotSolver_Astar::conduct_Move(int robot_id, int sUNUSED(src_vrtx_id), int dest_vrtx_id, const sRobotArrangement &current_arrangement, sRobotArrangement &next_arrangement)
    {
	next_arrangement = current_arrangement;
	next_arrangement.move_Robot(robot_id, dest_vrtx_id);
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolver_HCAstar

    const sMultirobotSolver_HCAstar::ProductionMove sMultirobotSolver_HCAstar::UNDEFINED_PRODUCTION_MOVE = sMultirobotSolver_HCAstar::ProductionMove(NULL, sMultirobotSolution::UNDEFINED_MOVE);


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_HCAstar::ProductionMove::ProductionMove(ProductionMove *prev, sMultirobotSolution::Move move)
	: m_prev(prev)
	, m_move(move)
    {
	// nothing
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_HCAstar::StateRecord::StateRecord(int location_id, int time_step)
	: m_location_id(location_id)
	, m_time_step(time_step)
	, m_production_move(NULL)
    {
	// nothing
    }


    bool sMultirobotSolver_HCAstar::StateRecord::operator==(const StateRecord &state_record) const
    {
	return (m_time_step == state_record.m_time_step && m_location_id == state_record.m_location_id);
    }


    bool sMultirobotSolver_HCAstar::StateRecord::operator<(const StateRecord &state_record) const
    {
	return (m_time_step < state_record.m_time_step || (m_time_step == state_record.m_time_step && m_location_id < state_record.m_location_id));
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_HCAstar::ReservationRecord::ReservationRecord(int vertex_id, int time_step)
	: m_vertex_id(vertex_id)
	, m_time_step(time_step)
    {
	// nothing
    }


    bool sMultirobotSolver_HCAstar::ReservationRecord::operator==(const ReservationRecord &reservation_record) const
    {
	return (m_vertex_id == reservation_record.m_vertex_id && m_time_step == reservation_record.m_time_step);
    }


    bool sMultirobotSolver_HCAstar::ReservationRecord::operator<(const ReservationRecord &reservation_record) const
    {
	return (m_time_step < reservation_record.m_time_step || (m_time_step == reservation_record.m_time_step && m_vertex_id < reservation_record.m_vertex_id));
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_HCAstar::sMultirobotSolver_HCAstar()
	: m_heuristic(NULL)
    {
	// nothing
    }

    
    void sMultirobotSolver_HCAstar::setup_Solver(sAstarHeuristic *heuristic)
    {
	sMultirobotSolver::setup_Solver();
	m_heuristic = heuristic;
    }


    void sMultirobotSolver_HCAstar::setup_Instance(const sMultirobotInstance &multirobot_instance)
    {
	sMultirobotSolver::setup_Instance(multirobot_instance);
	m_heuristic->setup_Instance(multirobot_instance);

	int N_Robots = m_multirobot_instance.m_initial_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    m_reservation_Table.insert(ReservationRecord(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id), 0));
	}
    }


    bool sMultirobotSolver_HCAstar::solve_Instance(sMultirobotSolution &multirobot_solution)
    {
	int N_Robots = m_multirobot_instance.m_initial_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    if (!solve_Robot(robot_id, multirobot_solution))
	    {
		printf("Robot %d not solved\n", robot_id);
		return false;
	    }
	}
	return true;
    }


    bool sMultirobotSolver_HCAstar::solve_Robot(int robot_id, sMultirobotSolution &multirobot_solution)
    {
	m_open_state_Recs.clear();
	m_close_state_Recs.clear();
	m_expansion_Queue.clear();
	m_production_Moves.clear();

	StateRecord first_state_rec(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id), 0);
	first_state_rec.m_cost = m_heuristic->calc_Heuristic(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id), m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id));
	StateRecords_set::iterator first_state_iter = m_open_state_Recs.insert(first_state_rec).first;
	m_expansion_Queue.insert(StateRecords_multimap::value_type(first_state_rec.m_cost, first_state_iter));

	while (!m_open_state_Recs.empty())
	{
	    StateRecords_multimap::iterator expand_record = m_expansion_Queue.begin();

	    if (expand_State(robot_id, *expand_record->second, m_reservation_Table, multirobot_solution))
	    {
		return true;
	    }
	    m_open_state_Recs.erase(expand_record->second);
	    m_expansion_Queue.erase(expand_record);
	}

	return false;
    }


    bool sMultirobotSolver_HCAstar::expand_State(int robot_id, const StateRecord &current_state_rec, ReservationTable_set &reservation_Table, sMultirobotSolution &multirobot_solution)
    {
	printf("Expanding state:%d,%d\n", current_state_rec.m_time_step, current_state_rec.m_location_id);
	sVertex *robot_location = m_multirobot_instance.m_environment.get_Vertex(current_state_rec.m_location_id);

	for (sVertex::Neighbors_list::const_iterator neighbor = robot_location->m_Neighbors.begin(); neighbor != robot_location->m_Neighbors.end(); ++neighbor)
	{
	    if (   reservation_Table.find(ReservationRecord((*neighbor)->m_target->m_id, current_state_rec.m_time_step)) == reservation_Table.end()
		&& reservation_Table.find(ReservationRecord((*neighbor)->m_target->m_id, current_state_rec.m_time_step + 1)) == reservation_Table.end())
	    {
		StateRecord next_state_rec((*neighbor)->m_target->m_id, current_state_rec.m_time_step + 1);
		
		if (m_open_state_Recs.find(next_state_rec) == m_open_state_Recs.end() && m_close_state_Recs.find(next_state_rec) == m_close_state_Recs.end())
		{
		    m_production_Moves.push_back(ProductionMove(current_state_rec.m_production_move, sMultirobotSolution::Move(robot_id, current_state_rec.m_location_id, (*neighbor)->m_target->m_id)));
		    next_state_rec.m_production_move = &m_production_Moves.back();
		    next_state_rec.m_cost = m_heuristic->calc_Heuristic((*neighbor)->m_target->m_id, m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id)) + current_state_rec.m_cost;	    
		    if ((*neighbor)->m_target->m_id == m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id))
		    {
			collect_Subsolution(next_state_rec, multirobot_solution, reservation_Table);
			return true;
		    }																						
		    StateRecords_set::iterator next_state_iter = m_open_state_Recs.insert(next_state_rec).first;		    
		    m_expansion_Queue.insert(StateRecords_multimap::value_type(next_state_rec.m_cost, next_state_iter));
		}
	    }
	}
	if (   reservation_Table.find(ReservationRecord(current_state_rec.m_location_id, current_state_rec.m_time_step)) == reservation_Table.end()
	    && reservation_Table.find(ReservationRecord(current_state_rec.m_location_id, current_state_rec.m_time_step + 1)) == reservation_Table.end())
	{
	    StateRecord next_state_rec(current_state_rec.m_location_id, current_state_rec.m_time_step + 1);
	    
	    if (m_open_state_Recs.find(next_state_rec) == m_open_state_Recs.end() && m_close_state_Recs.find(next_state_rec) == m_close_state_Recs.end())
	    {
		m_production_Moves.push_back(ProductionMove(current_state_rec.m_production_move, sMultirobotSolution::Move(robot_id, current_state_rec.m_location_id, current_state_rec.m_location_id)));
		next_state_rec.m_production_move = &m_production_Moves.back();
		next_state_rec.m_cost = m_heuristic->calc_Heuristic(current_state_rec.m_location_id, m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id)) + current_state_rec.m_cost;
		
		StateRecords_set::iterator next_state_iter = m_open_state_Recs.insert(next_state_rec).first;		    
		m_expansion_Queue.insert(StateRecords_multimap::value_type(next_state_rec.m_cost, next_state_iter));
	    }
	}
	m_close_state_Recs.insert(current_state_rec);

	return false;
    }


    int sMultirobotSolver_HCAstar::collect_Subsolution(const StateRecord &current_state_rec, sMultirobotSolution &multirobot_solution, ReservationTable_set &reservation_Table)
    {
	ProductionMove *production_move = current_state_rec.m_production_move;
	vector<ProductionMove*> production_Moves;

	for (; production_move != NULL; production_move = production_move->m_prev)
	{
	    production_Moves.push_back(production_move);
	}

	int time = 0;
	for (vector<ProductionMove*>::const_reverse_iterator prod_move = production_Moves.rbegin(); prod_move != production_Moves.rend(); ++prod_move, ++time)
	{
	    if ((*prod_move)->m_move.m_src_vrtx_id != (*prod_move)->m_move.m_dest_vrtx_id)
	    {
		multirobot_solution.add_Move(time, (*prod_move)->m_move);
		printf("%d-%d\n", (*prod_move)->m_move.m_src_vrtx_id, time);
		reservation_Table.insert(ReservationRecord((*prod_move)->m_move.m_src_vrtx_id, time));
	    }
	    else
	    {
		printf("Waiting...\n");
		printf("%d-%d\n", (*prod_move)->m_move.m_src_vrtx_id, time);
		reservation_Table.insert(ReservationRecord((*prod_move)->m_move.m_src_vrtx_id, time));
	    }
	}
	reservation_Table.insert(ReservationRecord((*production_Moves.begin())->m_move.m_dest_vrtx_id, time));
	printf("%d-%d\n", (*production_Moves.begin())->m_move.m_dest_vrtx_id, time);
	printf("Reservation table size:%ld\n", reservation_Table.size());

	return time;
    }


    void sMultirobotSolver_HCAstar::to_Screen(const ReservationTable_set &reservation_Table, const sString &indent) const
    {
	int last_time_step = -1;

	printf("%sReservation table [", indent.c_str());
	for (ReservationTable_set::const_iterator reservation_record = reservation_Table.begin(); reservation_record != reservation_Table.end(); ++reservation_record)
	{
	    if (last_time_step != reservation_record->m_time_step)
	    {
		printf("\n%s%s", indent.c_str(), sRELOC_INDENT.c_str());
		last_time_step = reservation_record->m_time_step;
	    }
	    printf("(%d,%d) ", reservation_record->m_time_step, reservation_record->m_vertex_id);
	}
	printf("\n%s]\n", indent.c_str());
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolver_WHCAstar

    const sMultirobotSolver_WHCAstar::ProductionMove sMultirobotSolver_WHCAstar::UNDEFINED_PRODUCTION_MOVE = sMultirobotSolver_WHCAstar::ProductionMove(NULL, sMultirobotSolution::UNDEFINED_MOVE);


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_WHCAstar::ProductionMove::ProductionMove(ProductionMove *prev, sMultirobotSolution::Move move)
	: m_prev(prev)
	, m_move(move)
    {
	// nothing
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_WHCAstar::StateRecord::StateRecord(int location_id, int time_step)
	: m_location_id(location_id)
	, m_time_step(time_step)
	, m_production_move(NULL)
    {
	// nothing
    }


    bool sMultirobotSolver_WHCAstar::StateRecord::operator==(const StateRecord &state_record) const
    {
	return (m_time_step == state_record.m_time_step && m_location_id == state_record.m_location_id);
    }


    bool sMultirobotSolver_WHCAstar::StateRecord::operator<(const StateRecord &state_record) const
    {
	return (m_time_step < state_record.m_time_step || (m_time_step == state_record.m_time_step && m_location_id < state_record.m_location_id));
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_WHCAstar::ReservationRecord::ReservationRecord(int vertex_id, int time_step)
	: m_vertex_id(vertex_id)
	, m_time_step(time_step)
    {
	// nothing
    }


    bool sMultirobotSolver_WHCAstar::ReservationRecord::operator==(const ReservationRecord &reservation_record) const
    {
	return (m_vertex_id == reservation_record.m_vertex_id && m_time_step == reservation_record.m_time_step);
    }


    bool sMultirobotSolver_WHCAstar::ReservationRecord::operator<(const ReservationRecord &reservation_record) const
    {
	return (m_time_step < reservation_record.m_time_step || (m_time_step == reservation_record.m_time_step && m_vertex_id < reservation_record.m_vertex_id));
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolver_WHCAstar::sMultirobotSolver_WHCAstar()
	: m_heuristic(NULL)
    {
	// nothing
    }

    
    void sMultirobotSolver_WHCAstar::setup_Solver(sAstarHeuristic *heuristic, int window_size, int max_N_Iterations)
    {
	sMultirobotSolver::setup_Solver();
	m_heuristic = heuristic;
	m_max_N_Iterations = max_N_Iterations;
	m_window_size = window_size;
    }


    void sMultirobotSolver_WHCAstar::setup_Instance(const sMultirobotInstance &multirobot_instance)
    {
	sMultirobotSolver::setup_Instance(multirobot_instance);
	m_heuristic->setup_Instance(multirobot_instance);
    }


    bool sMultirobotSolver_WHCAstar::solve_Instance(sMultirobotSolution &multirobot_solution)
    {
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("WHCA*_solving");
	}
        #endif

	int start_step = 0;

	for (int iteration = 0; iteration < m_max_N_Iterations; ++iteration)
	{
	    sRobotArrangement intermediate_arrangement;
	    sMultirobotSolution sub_solution;

	    if (!solve_Window(sub_solution))
	    {
                #ifdef sSTATISTICS
		{
		    s_GlobalPhaseStatistics.leave_Phase();
		}
                #endif
		return false;
	    }
	    int execution_N_Steps = sMIN(m_window_size / 2, sub_solution.get_StepCount());
	    sub_solution.execute_Solution(m_multirobot_instance.m_initial_arrangement, intermediate_arrangement, execution_N_Steps);

	    int solved_Robot_cnt = 0;
	    int N_Robots = m_multirobot_instance.m_initial_arrangement.get_RobotCount();
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		if (intermediate_arrangement.get_RobotLocation(robot_id) == m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id))
		{
		    ++solved_Robot_cnt;
		}
	    }	    

	    m_multirobot_instance.m_initial_arrangement = intermediate_arrangement;
	    sMultirobotSolution partial_sub_solution = sMultirobotSolution(start_step, sub_solution.extract_Subsolution(0, execution_N_Steps - 1));
	    multirobot_solution = sMultirobotSolution(multirobot_solution, partial_sub_solution);

	    start_step += execution_N_Steps;

	    if (solved_Robot_cnt >= N_Robots)
	    {
                #ifdef sSTATISTICS
		{
		    s_GlobalPhaseStatistics.leave_Phase();
		}
                #endif
		return true;
	    }
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return false;
    }


    bool sMultirobotSolver_WHCAstar::solve_Window(sMultirobotSolution &multirobot_solution)
    {
	m_reservation_Table.clear();

	int N_Robots = m_multirobot_instance.m_initial_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    m_reservation_Table.insert(ReservationRecord(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id), 0));
	}

	RobotPriority_multimap priority_Queue;
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int distance = m_heuristic->calc_Heuristic(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id),
						       m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id));
	    priority_Queue.insert(RobotPriority_multimap::value_type(distance, robot_id));
	}
	for (RobotPriority_multimap::const_iterator queue_record = priority_Queue.begin(); queue_record != priority_Queue.end(); ++queue_record)
	{
	    if (!solve_Robot(queue_record->second, multirobot_solution))
	    {
		return false;
	    }
	}
	return true;
    }


    bool sMultirobotSolver_WHCAstar::solve_Robot(int robot_id, sMultirobotSolution &multirobot_solution)
    {
	m_open_state_Recs.clear();
	m_close_state_Recs.clear();
	m_expansion_Queue.clear();
	m_production_Moves.clear();

	StateRecord first_state_rec(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id), 0);
	first_state_rec.m_cost = m_heuristic->calc_Heuristic(m_multirobot_instance.m_initial_arrangement.get_RobotLocation(robot_id), m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id));
	StateRecords_set::iterator first_state_iter = m_open_state_Recs.insert(first_state_rec).first;
	m_expansion_Queue.insert(StateRecords_multimap::value_type(first_state_rec.m_cost, first_state_iter));

        #ifdef sSTATISTICS
	{
	    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_search_Steps;
	}
        #endif

	while (!m_open_state_Recs.empty())
	{
	    StateRecords_multimap::iterator expand_record = m_expansion_Queue.begin();

	    if (expand_record->second->m_time_step >= m_window_size)
	    {
		collect_Subsolution(*expand_record->second, multirobot_solution, m_reservation_Table);
		return true;
	    }
	    else
	    {
		expand_State(robot_id, *expand_record->second, m_reservation_Table);
	    }
	    m_open_state_Recs.erase(expand_record->second);
	    m_expansion_Queue.erase(expand_record);
	}
	return false;
    }


    bool sMultirobotSolver_WHCAstar::expand_State(int robot_id, const StateRecord &current_state_rec, const ReservationTable_set &reservation_Table)
    {
	bool expanded = false;
	sVertex *robot_location = m_multirobot_instance.m_environment.get_Vertex(current_state_rec.m_location_id);

	for (sVertex::Neighbors_list::const_iterator neighbor = robot_location->m_Neighbors.begin(); neighbor != robot_location->m_Neighbors.end(); ++neighbor)
	{
	    if (   reservation_Table.find(ReservationRecord((*neighbor)->m_target->m_id, current_state_rec.m_time_step)) == reservation_Table.end()
		&& reservation_Table.find(ReservationRecord((*neighbor)->m_target->m_id, current_state_rec.m_time_step + 1)) == reservation_Table.end()
		&& reservation_Table.find(ReservationRecord((*neighbor)->m_target->m_id, current_state_rec.m_time_step + 2)) == reservation_Table.end())
	    {
		StateRecord next_state_rec((*neighbor)->m_target->m_id, current_state_rec.m_time_step + 1);
		
		if (m_open_state_Recs.find(next_state_rec) == m_open_state_Recs.end() && m_close_state_Recs.find(next_state_rec) == m_close_state_Recs.end())
		{
		    m_production_Moves.push_back(ProductionMove(current_state_rec.m_production_move, sMultirobotSolution::Move(robot_id, current_state_rec.m_location_id, (*neighbor)->m_target->m_id)));
		    next_state_rec.m_production_move = &m_production_Moves.back();
		    next_state_rec.m_cost = m_heuristic->calc_Heuristic((*neighbor)->m_target->m_id, m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id)) + current_state_rec.m_cost;		    
		    StateRecords_set::iterator next_state_iter = m_open_state_Recs.insert(next_state_rec).first;		    
		    m_expansion_Queue.insert(StateRecords_multimap::value_type(next_state_rec.m_cost, next_state_iter));
		    expanded = true;
		}
	    }
            #ifdef sSTATISTICS
	    {
		++s_GlobalPhaseStatistics.get_CurrentPhase().m_search_Steps;
	    }
            #endif
	}
	if (   reservation_Table.find(ReservationRecord(current_state_rec.m_location_id, current_state_rec.m_time_step)) == reservation_Table.end()
	    && reservation_Table.find(ReservationRecord(current_state_rec.m_location_id, current_state_rec.m_time_step + 1)) == reservation_Table.end()
	    && reservation_Table.find(ReservationRecord(current_state_rec.m_location_id, current_state_rec.m_time_step + 2)) == reservation_Table.end())
	{
	    StateRecord next_state_rec(current_state_rec.m_location_id, current_state_rec.m_time_step + 1);
	    
	    if (m_open_state_Recs.find(next_state_rec) == m_open_state_Recs.end() && m_close_state_Recs.find(next_state_rec) == m_close_state_Recs.end())
	    {
		m_production_Moves.push_back(ProductionMove(current_state_rec.m_production_move, sMultirobotSolution::Move(robot_id, current_state_rec.m_location_id, current_state_rec.m_location_id)));
		next_state_rec.m_production_move = &m_production_Moves.back();
		next_state_rec.m_cost = m_heuristic->calc_Heuristic(current_state_rec.m_location_id, m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id)) + current_state_rec.m_cost;
	
		StateRecords_set::iterator next_state_iter = m_open_state_Recs.insert(next_state_rec).first;		    
		m_expansion_Queue.insert(StateRecords_multimap::value_type(next_state_rec.m_cost, next_state_iter));
		expanded = true;
	    }
	}
        #ifdef sSTATISTICS
	{
	    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_search_Steps;
	}
        #endif
	m_close_state_Recs.insert(current_state_rec);

	return expanded;	    
    }
    

    int sMultirobotSolver_WHCAstar::collect_Subsolution(const StateRecord &current_state_rec, sMultirobotSolution &multirobot_solution, ReservationTable_set &reservation_Table)
    {
	ProductionMove *production_move = current_state_rec.m_production_move;
	vector<ProductionMove*> production_Moves;

	for (; production_move != NULL; production_move = production_move->m_prev)
	{
	    production_Moves.push_back(production_move);
	}

	int time = 0;
	for (vector<ProductionMove*>::const_reverse_iterator prod_move = production_Moves.rbegin(); prod_move != production_Moves.rend(); ++prod_move, ++time)
	{
	    if ((*prod_move)->m_move.m_src_vrtx_id != (*prod_move)->m_move.m_dest_vrtx_id)
	    {
		multirobot_solution.add_Move(time, (*prod_move)->m_move);
		reservation_Table.insert(ReservationRecord((*prod_move)->m_move.m_src_vrtx_id, time));
	    }
	    else
	    {
		reservation_Table.insert(ReservationRecord((*prod_move)->m_move.m_src_vrtx_id, time));
	    }
	}
	reservation_Table.insert(ReservationRecord((*production_Moves.begin())->m_move.m_dest_vrtx_id, time));

	return time;
    }


    void sMultirobotSolver_WHCAstar::to_Screen(const ReservationTable_set &reservation_Table, const sString &indent) const
    {
	int last_time_step = -1;

	printf("%sReservation table [", indent.c_str());
	for (ReservationTable_set::const_iterator reservation_record = reservation_Table.begin(); reservation_record != reservation_Table.end(); ++reservation_record)
	{
	    if (last_time_step != reservation_record->m_time_step)
	    {
		printf("\n%s%s", indent.c_str(), sRELOC_INDENT.c_str());
		last_time_step = reservation_record->m_time_step;
	    }
	    printf("(%d,%d) ", reservation_record->m_time_step, reservation_record->m_vertex_id);
	}
	printf("\n%s]\n", indent.c_str());
    }


/*----------------------------------------------------------------------------*/
// sMultirobotSolver_IDAstar

    sMultirobotSolver_IDAstar::sMultirobotSolver_IDAstar()
	: m_max_cost_limit(0),
	  m_finished_Robot_cnt(0)
    {
	// nothing
    }


    void sMultirobotSolver_IDAstar::setup_Solver(int max_cost_limit, sAstarHeuristic *heuristic)
    {
	sMultirobotSolver::setup_Solver();
	m_heuristic = heuristic;
	m_max_cost_limit = max_cost_limit;
    }


    void sMultirobotSolver_IDAstar::setup_Instance(const sMultirobotInstance &multirobot_instance)
    {
	sMultirobotSolver::setup_Instance(multirobot_instance);
	m_heuristic->setup_Instance(multirobot_instance);

	m_current_arrangement = multirobot_instance.m_initial_arrangement;

	m_finished_Robot_cnt = 0;

	int N_Robots = m_current_arrangement.get_RobotCount();
	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    if (m_current_arrangement.get_RobotLocation(robot_id) == m_multirobot_instance.m_goal_arrangement.get_RobotLocation(robot_id))
	    {
		++m_finished_Robot_cnt;
	    }
	}
    }

    
    bool sMultirobotSolver_IDAstar::solve_Instance(sMultirobotSolution &multirobot_solution)
    {
	for (int cost_limit = 0; m_max_cost_limit < 0 || cost_limit <= m_max_cost_limit;)
	{
	    int next_cost_limit;
	    bool solution_found = search_Goal(0, cost_limit, next_cost_limit, multirobot_solution);

	    if (solution_found)
	    {
		return true;
	    }

	    cost_limit = next_cost_limit;
	}

	return false;
    }


    bool sMultirobotSolver_IDAstar::search_Goal(int current_depth, int cost_limit, int &next_cost_limit, sMultirobotSolution &multirobot_solution)
    {
#ifdef sSTATISTICS
	static sUInt_64 search_Step_cnt = 0;
	++search_Step_cnt;
#endif    

#ifdef sVERBOSE	
	static sUInt_64 verbose_period = 65536;
	static sUInt_64 Verbose_cnt = 0;

	if (Verbose_cnt++ >= verbose_period)
	{
	    printf("Current depth/limit (steps): ");
	    for (int i = 0; i < cost_limit; ++i)
	    {
		printf(".");
	    }
	    printf(" %d/%d (%llu)\n", current_depth, cost_limit, search_Step_cnt);

	    Verbose_cnt = 0;
	    verbose_period *= 1.8;
	    ++verbose_period;
	}
#endif

	if (m_finished_Robot_cnt == m_current_arrangement.get_RobotCount())
	{
	    return true;
	}
	else
	{
	    next_cost_limit = INT_MAX;

	    int N_Robots = m_current_arrangement.get_RobotCount();
	    for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	    {
		int robot_vertex_id = m_current_arrangement.get_RobotLocation(robot_id);
		sVertex* robot_vertex = m_multirobot_instance.m_environment.get_Vertex(robot_vertex_id);
		
		for (sVertex::Neighbors_list::const_iterator neighbor = robot_vertex->m_Neighbors.begin(); neighbor != robot_vertex->m_Neighbors.end(); ++neighbor)
		{
		    int neighbor_vertex_id = (*neighbor)->m_target->m_id;
		    
		    if (m_current_arrangement.get_VertexOccupancy(neighbor_vertex_id) == sRobotArrangement::VACANT_VERTEX)
		    {
			conduct_Move(robot_id, robot_vertex_id, neighbor_vertex_id);
			int cost = current_depth + m_heuristic->calc_Heuristic(m_current_arrangement, m_multirobot_instance.m_goal_arrangement);
			
			if (cost <= cost_limit)
			{
			    int node_cost_limit;
			    bool result = search_Goal(current_depth + 1, cost_limit, node_cost_limit, multirobot_solution);
			    
			    if (result)
			    {
				multirobot_solution.add_Move(current_depth, sMultirobotSolution::Move(robot_id, robot_vertex_id, neighbor_vertex_id));
				return true;
			    }
			    next_cost_limit = sMIN(next_cost_limit, node_cost_limit);
			}
			else
			{
			    if (cost > cost_limit)
			    {
				next_cost_limit = sMIN(next_cost_limit, cost);
			    }
			}
			conduct_Move(robot_id, neighbor_vertex_id, robot_vertex_id);
		    }
		}
	    }
	    return false;
	}
	
	sASSERT(false);
	return false;
    }


    void sMultirobotSolver_IDAstar::conduct_Move(int robot_id, int src_vrtx_id, int dest_vrtx_id)
    {
	const sRobotArrangement &goal_arrangement = m_multirobot_instance.m_goal_arrangement;

	if (src_vrtx_id == goal_arrangement.get_RobotLocation(robot_id) && src_vrtx_id != dest_vrtx_id)
	{
	    --m_finished_Robot_cnt;
	}
	else
	{
	    if (dest_vrtx_id == goal_arrangement.get_RobotLocation(robot_id) && dest_vrtx_id != src_vrtx_id)
	    {
		++m_finished_Robot_cnt;
	    }
	}
	m_current_arrangement.move_Robot(robot_id, dest_vrtx_id);
    }


/*----------------------------------------------------------------------------*/

} // namespace sReloc
